""" Generic runtime controller. """

# -*- coding: utf-8 -*-
# clean
import importlib
from dataclasses import dataclass
from typing import Any, List

from dataclasses_json import dataclass_json

from hisim import utils
from hisim.component import ConfigBase, DisplayConfig
from hisim import component as cp
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class L1Config(ConfigBase):
    """L1 Runtime Config."""

    building_name: str
    name: str
    source_weight: int
    min_operation_time_in_seconds: int
    min_idle_time_in_seconds: int

    @staticmethod
    def get_default_config(
        name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Default config."""
        config = L1Config(
            building_name=building_name,
            name="RuntimeController_" + name,
            source_weight=1,
            min_operation_time_in_seconds=3600,
            min_idle_time_in_seconds=900,
        )
        return config

    @staticmethod
    def get_default_config_heatpump(
        name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config for heat pumps."""
        config = L1Config(
            building_name=building_name,
            name="L1RuntimeController" + name,
            source_weight=1,
            min_operation_time_in_seconds=3600 * 3,
            min_idle_time_in_seconds=3600,
        )
        return config


class L1GenericRuntimeControllerState:
    """The data class saves the state of the controller."""

    def __init__(
        self,
        on_off: int,
        activation_time_step: int = 0,
        deactivation_time_step: int = 0,
    ) -> None:
        """Initializes the data class."""
        self.on_off: int = on_off
        self.activation_time_step: int = activation_time_step
        self.deactivation_time_step: int = deactivation_time_step

    def clone(self) -> Any:
        """Generates a new state."""
        return L1GenericRuntimeControllerState(
            activation_time_step=self.activation_time_step,
            on_off=self.on_off,
            deactivation_time_step=self.deactivation_time_step,
        )

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def activate(self, timestep: int) -> None:
        """Activates the controller."""
        self.on_off = 1
        self.activation_time_step = timestep

    def deactivate(self, timestep: int) -> None:
        """Activates the controller."""
        self.on_off = 0
        self.deactivation_time_step = timestep


class L1GenericRuntimeController(cp.Component):
    """L1 Heat Pump Controller.

    It takes care of the operation of the heat pump only in terms of running times.
    It gets inputs from an L2-heat controller.

    Parameters
    ----------
    min_running_time: int, optional
        Minimal running time of device, in seconds. The default is 3600 seconds.
    min_idle_time : int, optional
        Minimal off time of device, in seconds. The default is 900 seconds.
    source_weight : int, optional
        Weight of component, relevant if there is more than one component of same type, defines hierachy in control. The default is 1.
    component type : str, optional
        Name of component to be controlled

    """

    # Inputs
    L2DeviceSignal = "l2_DeviceSignal"

    # Outputs
    L1DeviceSignal = "L1DeviceSignal"
    L1RunTimeSignal = "L1RunTimeSignal"

    # Similar components to connect to:
    # 1. building_name
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: L1Config,
        my_display_config: DisplayConfig = DisplayConfig(),
    ) -> None:
        """Initializes the controller."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.config = config
        self.name = config.name
        self.source_weight = config.source_weight
        self.minimum_runtime_in_timesteps = int(
            config.min_operation_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep
        )
        self.minimum_resting_time_in_timesteps = int(
            config.min_idle_time_in_seconds / self.my_simulation_parameters.seconds_per_timestep
        )

        self.state = L1GenericRuntimeControllerState(0, 0, 0)
        self.previous_state = self.state.clone()
        # add inputs
        self.l2_device_signal_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.L2DeviceSignal,
            LoadTypes.ON_OFF,
            Units.BINARY,
            mandatory=True,
        )
        self.add_default_connections(self.get_default_connections_from_controller_l2_generic_heat_clever_simple())

        # add outputs
        self.l1_device_signal_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.L1DeviceSignal, LoadTypes.ON_OFF, Units.BINARY
        )
        self.l1_runtime_signal: cp.ComponentOutput = self.add_output(
            self.component_name, self.L1RunTimeSignal, LoadTypes.ANY, Units.ANY
        )

    def get_default_connections_from_controller_l2_generic_heat_clever_simple(
        self,
    ) -> List[cp.ComponentConnection]:
        """Makes default connections to l2 smart controllers."""

        # use importlib for importing the other component in order to avoid circular-import errors
        component_module_name = "hisim.components.controller_l2_generic_heat_clever_simple"
        component_module = importlib.import_module(name=component_module_name)
        component_class = getattr(component_module, "L2HeatSmartController")
        connections = []
        controller_classname = component_class.get_classname()
        connections.append(
            cp.ComponentConnection(
                L1GenericRuntimeController.L2DeviceSignal,
                controller_classname,
                component_class.L2DeviceSignal,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the current state for the next timestep."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the previous state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """For checking for problems."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Main simulation function."""
        # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
        if force_convergence:
            # states are saved after each timestep, outputs after each iteration
            # outputs have to be in line with states, so if convergence is forced outputs are aligned to state.
            stsv.set_output_value(self.l1_device_signal_channel, self.state.on_off)
            return

        l2_devicesignal = stsv.get_input_value(self.l2_device_signal_channel)

        # return device on if minimum operation time is not fulfilled and device was on in previous state
        if self.state.on_off == 1 and self.state.activation_time_step + self.minimum_runtime_in_timesteps >= timestep:
            # mandatory on, minimum runtime not reached
            self.state.on_off = 1
            pass
        elif (
            self.state.on_off == 0
            and self.state.deactivation_time_step + self.minimum_resting_time_in_timesteps >= timestep
        ):
            self.state.on_off = 0
        # check signal from l2 and turn on or off if it is necesary
        else:
            if l2_devicesignal == 0 and self.state.on_off == 1:
                self.state.deactivate(timestep)
            elif l2_devicesignal == 1 and self.state.on_off == 0:
                self.state.activate(timestep)
        stsv.set_output_value(self.l1_device_signal_channel, self.state.on_off)

    def write_to_report(self) -> List[str]:
        """Writes config to report."""
        lines: List[str] = []
        lines.append("Generic Controller L1: " + self.component_name)
        lines.extend(self.config.get_string_dict())
        return lines



""" Iterative Energy Surplus Controller For Districts.

It received the electricity consumption
of all components and the PV production. According to the balance it
sends activation/deactivation signals to components.
The component with the lowest source weight is activated first.
"""

# clean
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, List, Tuple, Union
from collections import OrderedDict
from dataclasses_json import dataclass_json
import pandas as pd
from hisim import log
from hisim import component as cp
from hisim import dynamic_component
from hisim import loadtypes as lt
from hisim import utils
from hisim.component import ComponentInput, ComponentOutput
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiTagEnumClass, KpiHelperClass
from hisim.components import (
    more_advanced_heat_pump_hplib,
    advanced_heat_pump_hplib,
    generic_heat_pump_modular,
    loadprofilegenerator_utsp_connector,
)

__authors__ = ""
__copyright__ = ""
__credits__ = [""]
__license__ = ""
__version__ = ""
__maintainer__ = " "
__email__ = ""
__status__ = ""


class EMSControlStrategy(IntEnum):
    """Set Control Strategy of EMS."""

    BUILDING_OPTIMIZEOWNCONSUMPTION_ITERATIV = 1
    BUILDING_OPTIMIZEOWNCONSUMPTION_PARALLEL = 2
    DISTRICT_OPTIMIZECONSUMPTION_PARALLEL = 3


@dataclass_json
@dataclass
class EMSDistrictConfig(cp.ConfigBase):
    """L1 Controller Config."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return L2GenericDistrictEnergyManagementSystem.get_full_classname()

    building_name: str
    #: name of the device
    name: str
    # control strategy, more or less obsolete because only "optimize_own_consumption_interativ" is used at the moment.
    strategy: Union[EMSControlStrategy, int]
    # limit for peak shaving option, more or less obsolete because only "optimize_own_consumption_interativ" is used at the moment.
    limit_to_shave: float
    # increase building set temperatures for heating when PV surplus is available.
    # Must be smaller than difference of set_heating_temperature and set_cooling_temperature
    building_indoor_temperature_offset_value: float
    # increase in dhw buffer set temperatures when PV surplus is available for heating
    domestic_hot_water_storage_temperature_offset_value: float
    # increase in SimpleHotWaterStorage set temperatures when PV surplus is available for heating
    space_heating_water_storage_temperature_offset_value: float

    @classmethod
    def get_default_config_ems(
        cls,
        strategy: Union[EMSControlStrategy, int] = EMSControlStrategy.BUILDING_OPTIMIZEOWNCONSUMPTION_ITERATIV,
        building_name: str = "BUI1",
    ) -> "EMSDistrictConfig":
        """Default Config for Energy Management System."""
        config = EMSDistrictConfig(
            building_name=building_name,
            name="L2EMSElectricityController",
            strategy=strategy,
            limit_to_shave=0,
            building_indoor_temperature_offset_value=2,
            domestic_hot_water_storage_temperature_offset_value=20,
            space_heating_water_storage_temperature_offset_value=25,
        )
        return config


class EMSDistrictState:
    """Saves the state of the Energy Management System."""

    def __init__(
        self,
        production: float,
        consumption_uncontrolled: float,
        consumption_ems_controlled: float,
    ) -> None:
        """Initialize the heat pump controller state."""
        self.production_in_watt = production
        self.consumption_uncontrolled_in_watt = consumption_uncontrolled
        self.consumption_ems_controlled_in_watt = consumption_ems_controlled

    def clone(self) -> "EMSDistrictState":
        """Copy EMSState efficiently."""
        return EMSDistrictState(
            production=self.production_in_watt,
            consumption_uncontrolled=self.consumption_uncontrolled_in_watt,
            consumption_ems_controlled=self.consumption_ems_controlled_in_watt,
        )


class L2GenericDistrictEnergyManagementSystem(dynamic_component.DynamicComponent):
    """Surplus electricity controller - time step based.

    Iteratively goes through connected inputs by hierachy of
    source weights of inputs and passes available surplus
    electricity to each device. Needs to be configured with
    dynamic In- and Outputs.

    Recognises production of any component when dynamic input
    is labeled with the flag "CONSUMPTION" and the
    related source weight is set to 999.

    Recognised non controllable consumption of any component
    when dynamic input is labeld with the flag
    "ELECTRICITY_CONSUMPTION_UNCONTROLLED" and the related
    source weight is set to 999.

    For each component, which should receive signals from the
    EMS, the EMS needs to be connected with one dynamic input
    with the tag "ELECTRICITY_CONSUMPTION_EMS_CONTROLLED" and
    the source weight of the related component. This signal
    reflects the real consumption/production of the device,
    which is needed to update the energy balance in the EMS.
    In addition, the EMS needs to be connected with one dynamic
    output with the tag "ELECTRICITY_TARGET" with the
    source weight of the related component. This signal sends
    information on the available surplus electricity to the
    component, which receives signals from the EMS.

    """

    # Inputs
    ElectricityToElectrolyzerUnused = "ElectricityToElectrolyzerUnused"
    SurplusUnusedFromDistrictEMS = "SurplusUnusedFromDistrictEMS"

    # Outputs
    ElectricityToElectrolyzerTarget = "ElectricityToElectrolyzerTarget"

    TotalElectricityToOrFromGrid = "TotalElectricityToOrFromGrid"
    TotalElectricityToGrid = "TotalElectricityToGrid"
    TotalElectricityFromGrid = "TotalElectricityFromGrid"
    TotalElectricityConsumption = "TotalElectricityConsumption"
    ElectricityProduction = "ElectricityProduction"
    TotalElectricityConsumptionEMSControlled = "TotalElectricityConsumptionEMSControlled"
    TotalElectricityConsumptionUnControlled = "TotalElectricityConsumptionUnControlled"
    BuildingIndoorTemperatureModifier = "BuildingIndoorTemperatureModifier"  # connect to HDS controller and Building
    DomesticHotWaterStorageTemperatureModifier = (
        "DomesticHotWaterStorageTemperatureModifier"  # used for L1HeatPumpController  # Todo: change name?
    )
    SpaceHeatingWaterStorageTemperatureModifier = (
        "SpaceHeatingWaterStorageTemperatureModifier"  # used for HeatPumpHplibController
    )
    SurplusUnusedFromDistrictEMSOutput = "SurplusUnusedFromDistrictEMSOutput"
    SurplusUnusedFromBuildingEMSOutput = "SurplusUnusedFromBuildingEMSOutput"

    CheckPeakShaving = "CheckPeakShaving"

    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: EMSDistrictConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ):
        """Initializes."""
        self.my_component_inputs: List[dynamic_component.DynamicConnectionInput] = []
        self.my_component_outputs: List[dynamic_component.DynamicConnectionOutput] = []
        self.ems_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_component_inputs=self.my_component_inputs,
            my_component_outputs=self.my_component_outputs,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.state = EMSDistrictState(production=0, consumption_uncontrolled=0, consumption_ems_controlled=0)
        self.previous_state = self.state.clone()

        self.component_types_sorted: List[lt.ComponentType] = []
        self.inputs_sorted: List[ComponentInput] = []
        self.source_weights_sorted: List[int] = []
        self.outputs_sorted: List[ComponentOutput] = []
        self.production_inputs: List[ComponentInput] = []
        self.consumption_uncontrolled_inputs: List[ComponentInput] = []
        self.consumption_ems_controlled_inputs: List[ComponentInput] = []
        self.components_parameters: dict = {}
        self.component_and_number_of_same_source_weights: dict = {}

        self.mode: Any
        self.strategy = self.ems_config.strategy
        self.limit_to_shave = self.ems_config.limit_to_shave
        self.building_indoor_temperature_offset_value = self.ems_config.building_indoor_temperature_offset_value
        self.domestic_hot_water_storage_temperature_offset_value = (
            self.ems_config.domestic_hot_water_storage_temperature_offset_value
        )
        self.space_heating_water_storage_temperature_offset_value = (
            self.ems_config.space_heating_water_storage_temperature_offset_value
        )

        # Inputs
        self.electricity_to_electrolyzer_unused: cp.ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.ElectricityToElectrolyzerUnused,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            mandatory=False,
        )

        self.surplus_electricity_unused_to_building_ems_from_district_ems: cp.ComponentInput = self.add_input(
            object_name=self.component_name,
            field_name=self.SurplusUnusedFromDistrictEMS,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            mandatory=False,
        )

        # Outputs
        self.total_electricity_to_or_from_grid: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TotalElectricityToOrFromGrid,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.TotalElectricityToOrFromGrid} will follow.",
        )

        self.total_electricity_to_grid: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TotalElectricityToGrid,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.TotalElectricityToGrid} will follow.",
        )

        self.total_electricity_from_grid: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TotalElectricityFromGrid,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.TotalElectricityFromGrid} will follow.",
        )

        self.total_electricity_consumption_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TotalElectricityConsumption,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.TotalElectricityConsumption} will follow.",
            postprocessing_flag=(
                [
                    lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED,
                    lt.ComponentType.BUILDINGS,
                    lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
                ]
                if any(word in config.building_name for word in lt.DistrictNames)
                else []
            ),
        )

        self.total_electricity_consumption_ems_controlled_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TotalElectricityConsumptionEMSControlled,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.TotalElectricityConsumptionEMSControlled} will follow.",
        )

        self.electricity_production_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityProduction,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityProduction} will follow.",
        )

        self.total_electricity_consumption_uncontrolled_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.TotalElectricityConsumptionUnControlled,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.TotalElectricityConsumptionUnControlled} will follow.",
        )

        self.building_indoor_temperature_modifier: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.BuildingIndoorTemperatureModifier,
            load_type=lt.LoadTypes.TEMPERATURE,
            unit=lt.Units.CELSIUS,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.BuildingIndoorTemperatureModifier} will follow.",
        )

        self.domestic_hot_water_storage_temperature_modifier: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.DomesticHotWaterStorageTemperatureModifier,
            load_type=lt.LoadTypes.TEMPERATURE,
            unit=lt.Units.CELSIUS,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.DomesticHotWaterStorageTemperatureModifier} will follow.",
        )

        self.space_heating_water_storage_temperature_modifier: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.SpaceHeatingWaterStorageTemperatureModifier,
            load_type=lt.LoadTypes.TEMPERATURE,
            unit=lt.Units.CELSIUS,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.SpaceHeatingWaterStorageTemperatureModifier} will follow.",
        )

        self.check_peak_shaving: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.CheckPeakShaving,
            load_type=lt.LoadTypes.ANY,
            unit=lt.Units.ANY,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.CheckPeakShaving} will follow.",
        )

        self.surplus_electricity_unused_to_building_ems_from_district_ems_output: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.SurplusUnusedFromDistrictEMSOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.SurplusUnusedFromDistrictEMSOutput} will follow.",
        )
        if any(word in config.building_name for word in lt.DistrictNames):
            self.surplus_electricity_unused_to_district_ems_from_building_ems_output: cp.ComponentOutput = self.add_output(
                object_name=self.component_name,
                field_name=self.SurplusUnusedFromBuildingEMSOutput,
                load_type=lt.LoadTypes.ELECTRICITY,
                unit=lt.Units.WATT,
                sankey_flow_direction=False,
                output_description=f"here a description for {self.SurplusUnusedFromBuildingEMSOutput} will follow.",
                postprocessing_flag=(
                    [
                        lt.InandOutputType.ELECTRICITY_PRODUCTION,
                        lt.ComponentType.BUILDINGS,
                        lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
                    ]
                    if any(word in config.building_name for word in lt.DistrictNames)
                    else []
                ),
            )

        self.add_dynamic_default_connections(self.get_default_connections_from_utsp_occupancy())
        self.add_dynamic_default_connections(self.get_default_connections_from_pv_system())
        #  self.add_dynamic_default_connections(self.get_default_connections_from_more_advanced_heat_pump())
        self.add_dynamic_default_connections(self.get_default_connections_from_dhw_heat_pump())
        self.add_dynamic_default_connections(self.get_default_connections_from_advanced_heat_pump())
        self.add_dynamic_default_connections(self.get_default_connections_from_advanced_battery())

    def get_default_connections_from_pv_system(
        self,
    ):
        """Get pv system default connections."""

        from hisim.components.generic_pv_system import PVSystem  # pylint: disable=import-outside-toplevel

        dynamic_connections = []
        pv_class_name = PVSystem.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=PVSystem,
                source_class_name=pv_class_name,
                source_component_field_name=PVSystem.ElectricityOutput,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[
                    lt.ComponentType.PV,
                    lt.InandOutputType.ELECTRICITY_PRODUCTION,
                ],
                source_weight=999,
            )
        )

        return dynamic_connections

    def get_default_connections_from_utsp_occupancy(
        self,
    ):
        """Get utsp occupancy default connections."""

        from hisim.components.loadprofilegenerator_utsp_connector import (  # pylint: disable=import-outside-toplevel
            UtspLpgConnector,
        )

        dynamic_connections = []
        self.occupancy_class_name = UtspLpgConnector.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=UtspLpgConnector,
                source_class_name=self.occupancy_class_name,
                source_component_field_name=UtspLpgConnector.ElectricityOutput,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[lt.ComponentType.RESIDENTS, lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED],
                source_weight=1,
            )
        )
        self.add_component_output(
            source_output_name=f"ElectricityToOrFromGridOf{self.occupancy_class_name}_",
            source_tags=[
                lt.ComponentType.RESIDENTS,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=1,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for Occupancy. ",
        )
        return dynamic_connections

    # def get_default_connections_from_more_advanced_heat_pump(
    #         self,
    # ):
    #     """Get advanced heat pump default connections."""
    #
    #     from hisim.components.more_advanced_heat_pump_hplib import (  # pylint: disable=import-outside-toplevel
    #         MoreAdvancedHeatPumpHPLib,
    #     )
    #
    #     dynamic_connections = []
    #     self.more_advanced_heat_pump_class_name = MoreAdvancedHeatPumpHPLib.get_classname()
    #     dynamic_connections.append(
    #         dynamic_component.DynamicComponentConnection(
    #             source_component_class=MoreAdvancedHeatPumpHPLib,
    #             source_class_name=self.more_advanced_heat_pump_class_name,
    #             source_component_field_name=MoreAdvancedHeatPumpHPLib.ElectricalInputPowerSH,
    #             source_load_type=lt.LoadTypes.ELECTRICITY,
    #             source_unit=lt.Units.WATT,
    #             source_tags=[
    #                 lt.ComponentType.HEAT_PUMP_BUILDING,
    #                 lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
    #             ],
    #             source_weight=2,
    #         )
    #     )
    #     dynamic_connections.append(
    #         dynamic_component.DynamicComponentConnection(
    #             source_component_class=MoreAdvancedHeatPumpHPLib,
    #             source_class_name=self.more_advanced_heat_pump_class_name,
    #             source_component_field_name=MoreAdvancedHeatPumpHPLib.ElectricalInputPowerDHW,
    #             source_load_type=lt.LoadTypes.ELECTRICITY,
    #             source_unit=lt.Units.WATT,
    #             source_tags=[
    #                 lt.ComponentType.HEAT_PUMP_DHW,
    #                 lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
    #             ],
    #             source_weight=3,
    #         )
    #     )
    #     self.add_component_output(
    #         source_output_name=f"ElectricityToOrFromGridOfSH{self.more_advanced_heat_pump_class_name}_",
    #         source_tags=[
    #             lt.ComponentType.HEAT_PUMP_BUILDING,
    #             lt.InandOutputType.ELECTRICITY_TARGET,
    #         ],
    #         source_weight=2,
    #         source_load_type=lt.LoadTypes.ELECTRICITY,
    #         source_unit=lt.Units.WATT,
    #         output_description="Target electricity for Heating Heat Pump. ",
    #     )
    #     self.add_component_output(
    #         source_output_name=f"ElectricityToOrFromGridOfDHW{self.more_advanced_heat_pump_class_name}_",
    #         source_tags=[
    #             lt.ComponentType.HEAT_PUMP_DHW,
    #             lt.InandOutputType.ELECTRICITY_TARGET,
    #         ],
    #         source_weight=3,
    #         source_load_type=lt.LoadTypes.ELECTRICITY,
    #         source_unit=lt.Units.WATT,
    #         output_description="Target electricity for Heating Heat Pump. ",
    #     )
    #     return dynamic_connections

    def get_default_connections_from_advanced_heat_pump(
        self,
    ):
        """Get advanced heat pump default connections."""

        from hisim.components.advanced_heat_pump_hplib import HeatPumpHplib  # pylint: disable=import-outside-toplevel

        dynamic_connections = []
        self.advanced_heat_pump_class_name = HeatPumpHplib.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=HeatPumpHplib,
                source_class_name=self.advanced_heat_pump_class_name,
                source_component_field_name=HeatPumpHplib.ElectricalInputPower,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[
                    lt.ComponentType.HEAT_PUMP_BUILDING,
                    lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
                ],
                source_weight=2,
            )
        )
        self.add_component_output(
            source_output_name=f"ElectricityToOrFromGridOf{self.advanced_heat_pump_class_name}_",
            source_tags=[
                lt.ComponentType.HEAT_PUMP_BUILDING,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=2,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for Heating Heat Pump. ",
        )
        return dynamic_connections

    def get_default_connections_from_dhw_heat_pump(
        self,
    ):
        """Get dhw heat pump default connections."""

        from hisim.components.generic_heat_pump_modular import (  # pylint: disable=import-outside-toplevel
            ModularHeatPump,
        )

        dynamic_connections = []
        self.dhw_heat_pump_class_name = ModularHeatPump.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=ModularHeatPump,
                source_class_name=self.dhw_heat_pump_class_name,
                source_component_field_name=ModularHeatPump.ElectricityOutput,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[lt.ComponentType.HEAT_PUMP_DHW, lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED],
                source_weight=3,
            )
        )

        self.add_component_output(
            source_output_name=f"ElectricityToOrFromGridOf{self.dhw_heat_pump_class_name}_",
            source_tags=[
                lt.ComponentType.HEAT_PUMP_DHW,
                lt.InandOutputType.ELECTRICITY_TARGET,
            ],
            source_weight=3,
            source_load_type=lt.LoadTypes.ELECTRICITY,
            source_unit=lt.Units.WATT,
            output_description="Target electricity for dhw heat pump.",
        )
        return dynamic_connections

    def get_default_connections_from_advanced_battery(
        self,
    ):
        """Get advanced battery default connections."""

        from hisim.components.advanced_battery_bslib import Battery  # pylint: disable=import-outside-toplevel

        dynamic_connections = []
        advanced_battery_class_name = Battery.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=Battery,
                source_class_name=advanced_battery_class_name,
                source_component_field_name=Battery.AcBatteryPowerUsed,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[lt.ComponentType.BATTERY, lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED],
                source_weight=4,
            )
        )

        return dynamic_connections

    def sort_source_weights_and_components(
        self,
    ) -> Tuple[
        List[int],
        List[ComponentInput],
        List[lt.ComponentType],
        List[ComponentOutput],
        List[ComponentInput],
        List[ComponentInput],
        List[ComponentInput],
    ]:
        """Sorts dynamic Inputs and Outputs according to source weights."""
        inputs = [elem for elem in self.my_component_inputs if elem.source_weight != 999]

        source_tags = [elem.source_tags[0] for elem in inputs]
        source_weights = [elem.source_weight for elem in inputs]
        sortindex = sorted(range(len(source_weights)), key=lambda k: source_weights[k])
        source_weights = [source_weights[i] for i in sortindex]

        component_types_sorted = [source_tags[i] for i in sortindex]
        inputs_sorted = [getattr(self, inputs[i].source_component_class) for i in sortindex]
        outputs_sorted = []

        for ind, source_weight in enumerate(source_weights):
            outputs = self.get_all_dynamic_outputs(
                tags=[
                    component_types_sorted[ind],
                    lt.InandOutputType.ELECTRICITY_TARGET,
                ],
                weight_counter=source_weight,
            )

            for output in outputs:
                if output is not None:
                    outputs_sorted.append(output)
                else:
                    raise Exception("Dynamic input is not conncted to dynamic output")
        outputs_sorted = list(OrderedDict.fromkeys(outputs_sorted))
        production_inputs = self.get_dynamic_inputs(tags=[lt.InandOutputType.ELECTRICITY_PRODUCTION])
        consumption_uncontrolled_inputs = self.get_dynamic_inputs(
            tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED]
        )
        consumption_ems_controlled_inputs = self.get_dynamic_inputs(
            tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED]
        )
        return (
            source_weights,
            inputs_sorted,
            component_types_sorted,
            outputs_sorted,
            production_inputs,
            consumption_uncontrolled_inputs,
            consumption_ems_controlled_inputs,
        )

    def write_to_report(self):
        """Writes relevant information to report."""
        return self.ems_config.get_string_dict()

    def i_save_state(self) -> None:
        """Saves the state."""
        # abändern, siehe Storage
        self.previous_state = self.state

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks values."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates iteration of surplus controller."""
        if timestep == 0:
            (
                self.source_weights_sorted,
                self.inputs_sorted,
                self.component_types_sorted,
                self.outputs_sorted,
                self.production_inputs,
                self.consumption_uncontrolled_inputs,
                self.consumption_ems_controlled_inputs,
            ) = self.sort_source_weights_and_components()

            self.components_parameters = {}
            self.component_and_number_of_same_source_weights = {}

            for i, source_object_name in enumerate(self.inputs_sorted):
                self.components_parameters[source_object_name.field_name] = {
                    "input": self.inputs_sorted[i],
                    "output": self.outputs_sorted[i],
                    "component_type": self.component_types_sorted[i],
                    "weight": self.source_weights_sorted[i],
                }

            for params in self.components_parameters.values():
                weight = params["weight"]
                field_name = params["input"].field_name

                if weight in self.component_and_number_of_same_source_weights:
                    self.component_and_number_of_same_source_weights[weight]["count"] += 1
                    self.component_and_number_of_same_source_weights[weight]["components"].append(field_name)
                else:
                    self.component_and_number_of_same_source_weights[weight] = {
                        "count": 1,
                        "components": [field_name],
                    }

        district_electricity_surplus_unused = max(
            0.0, stsv.get_input_value(component_input=self.surplus_electricity_unused_to_building_ems_from_district_ems)
        )

        stsv.set_output_value(
            self.surplus_electricity_unused_to_building_ems_from_district_ems_output,
            district_electricity_surplus_unused,
        )

        # get total production and consumptions
        self.state.production_in_watt = (
            sum([stsv.get_input_value(component_input=elem) for elem in self.production_inputs])
            + district_electricity_surplus_unused
        )
        self.state.consumption_uncontrolled_in_watt = sum(
            [stsv.get_input_value(component_input=elem) for elem in self.consumption_uncontrolled_inputs]
        )
        self.state.consumption_ems_controlled_in_watt = sum(
            [stsv.get_input_value(component_input=elem) for elem in self.consumption_ems_controlled_inputs]
        )

        if any(word in self.config.building_name for word in lt.DistrictNames):
            production_inputs_from_buildings = self.get_dynamic_inputs(tags=[lt.InandOutputType.ELECTRICITY_PRODUCTION,
                                                                             lt.ComponentType.BUILDINGS])

            building_electricity_surplus_unused = (
                sum([stsv.get_input_value(component_input=elem) for elem in production_inputs_from_buildings]))

            stsv.set_output_value(
                self.surplus_electricity_unused_to_district_ems_from_building_ems_output,
                building_electricity_surplus_unused,
            )

        # Production of Electricity positve sign
        # Consumption of Electricity negative sign
        available_surplus_electricity_in_watt = (
            self.state.production_in_watt - self.state.consumption_uncontrolled_in_watt
        )

        if self.strategy == EMSControlStrategy.BUILDING_OPTIMIZEOWNCONSUMPTION_ITERATIV:
            available_surplus_electricity_in_watt = (
                self.distribute_available_surplus_electricity_to_building_components_iterative(
                    available_surplus_electricity_in_watt=available_surplus_electricity_in_watt,
                    stsv=stsv,
                    inputs_sorted=self.inputs_sorted,
                    component_types_sorted=self.component_types_sorted,
                    outputs_sorted=self.outputs_sorted,
                )
            )
            self.modify_set_temperatures_for_building_components_in_case_of_surplus_electricity_iterativ(
                available_surplus_electricity_in_watt=available_surplus_electricity_in_watt,
                stsv=stsv,
                inputs_sorted=self.inputs_sorted,
                component_types_sorted=self.component_types_sorted,
            )

        if self.strategy == EMSControlStrategy.BUILDING_OPTIMIZEOWNCONSUMPTION_PARALLEL:
            available_surplus_electricity_in_watt = (
                self.distribute_available_surplus_electricity_to_building_components_parallel(
                    components_parameters=self.components_parameters,
                    component_and_number_of_same_source_weights=self.component_and_number_of_same_source_weights,
                    available_surplus_electricity_in_watt=available_surplus_electricity_in_watt,
                    stsv=stsv,
                )
            )

        if self.strategy == EMSControlStrategy.DISTRICT_OPTIMIZECONSUMPTION_PARALLEL:
            available_surplus_electricity_in_watt = (
                self.distribute_available_surplus_electricity_to_building_components_parallel(
                    components_parameters=self.components_parameters,
                    component_and_number_of_same_source_weights=self.component_and_number_of_same_source_weights,
                    available_surplus_electricity_in_watt=available_surplus_electricity_in_watt,
                    stsv=stsv,
                )
            )

        stsv.set_output_value(self.total_electricity_to_or_from_grid, available_surplus_electricity_in_watt)
        stsv.set_output_value(
            self.total_electricity_to_grid,
            available_surplus_electricity_in_watt if available_surplus_electricity_in_watt > 0 else 0,
        )
        stsv.set_output_value(
            self.total_electricity_from_grid,
            -available_surplus_electricity_in_watt if available_surplus_electricity_in_watt < 0 else 0,
        )
        stsv.set_output_value(
            self.total_electricity_consumption_channel,
            self.state.consumption_uncontrolled_in_watt + self.state.consumption_ems_controlled_in_watt,
        )

        stsv.set_output_value(self.electricity_production_channel, self.state.production_in_watt)
        stsv.set_output_value(
            self.total_electricity_consumption_uncontrolled_channel, self.state.consumption_uncontrolled_in_watt
        )
        stsv.set_output_value(
            self.total_electricity_consumption_ems_controlled_channel, self.state.consumption_ems_controlled_in_watt
        )
        """
        elif self.strategy == "seasonal_storage":
            self.seasonal_storage(delta_demand=delta_demand, stsv=stsv)
        elif self.strategy == "peak_shaving_into_grid":
            self.peak_shaving_into_grid(delta_demand=delta_demand, limit_to_shave=limit_to_shave,stsv=stsv)
        elif self.strategy == "peak_shaving_from_grid":
            self.peak_shaving_from_grid(delta_demand=delta_demand, limit_to_shave=limit_to_shave,stsv=stsv)
        """

        # HEAT #
        # If comftortable temperature of building is to low heat with WarmWaterStorage the building
        # Solution with Control Signal Residence
        # not perfect solution!
        """
        if self.temperature_residence<self.min_comfortable_temperature_residence:
            #heat
            #here has to be added how "strong" HeatingWater Storage can be discharged
            #Working with upper boarder?
        elif self.temperature_residence > self.max_comfortable_temperature_residence:
            #cool
        elif self.temperature_residence>self.min_comfortable_temperature_residence and self.temperature_residence<self.max_comfortable_temperature_residence:
        """

    def distribute_available_surplus_electricity_to_building_components_iterative(
        self,
        available_surplus_electricity_in_watt: float,
        stsv: cp.SingleTimeStepValues,
        inputs_sorted: List[ComponentInput],
        component_types_sorted: List[lt.ComponentType],
        outputs_sorted: List[ComponentOutput],
    ) -> float:
        """Evaluates available surplus electricity component by component, iteratively, and sends updated signals back."""

        for index, single_input_sorted in enumerate(inputs_sorted):
            single_component_type_sorted = component_types_sorted[index]
            single_output_sorted = outputs_sorted[index]

            available_surplus_electricity_in_watt = self.control_electricity_iterative(
                available_surplus_electricity_in_watt=available_surplus_electricity_in_watt,
                stsv=stsv,
                current_component_type=single_component_type_sorted,
                current_input=single_input_sorted,
                current_output=single_output_sorted,
            )

        return available_surplus_electricity_in_watt

    def control_electricity_iterative(
        self,
        available_surplus_electricity_in_watt: float,
        stsv: cp.SingleTimeStepValues,
        current_component_type: lt.ComponentType,
        current_input: cp.ComponentInput,
        current_output: cp.ComponentOutput,
    ) -> float:
        """Calculates available surplus electricity.

        Subtracts the electricity consumption signal of the component from the previous iteration,
        and sends updated signal back.
        This function controls how surplus electricity is distributed and how much of each components'
        electricity need is covered onsite or from grid.
        """
        # get electricity demand from input component and substract from (or add to) available surplus electricity
        electricity_demand_from_current_input_component_in_watt = stsv.get_input_value(component_input=current_input)

        # if available_surplus_electricity > 0: electricity is fed into battery
        # if available_surplus_electricity < 0: electricity is taken from battery
        if current_component_type == lt.ComponentType.BATTERY:
            stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            # difference between what is fed into battery and what battery really used
            available_surplus_electricity_in_watt = (
                available_surplus_electricity_in_watt - electricity_demand_from_current_input_component_in_watt
            )

        # Building Tag, if district controller manage the hp in building --> contracting
        # Building has own ems which controll the temperatures of storages
        # District EMS controll how surplus is devided to buildings(ems)
        elif current_component_type == lt.ComponentType.SURPLUS_CONTROLLER_DISTRICT:
            if available_surplus_electricity_in_watt > 0:
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )
                stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            else:
                stsv.set_output_value(
                    output=current_output, value=electricity_demand_from_current_input_component_in_watt
                )
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )

        # these are electricity CONSUMERS
        elif current_component_type in [
            lt.ComponentType.RESIDENTS,
            lt.ComponentType.BUILDINGS,
            lt.ComponentType.ELECTROLYZER,
            lt.ComponentType.SMART_DEVICE,
            lt.ComponentType.CAR_BATTERY,
            lt.ComponentType.HEAT_PUMP_DHW,
            lt.ComponentType.HEAT_PUMP,
            lt.ComponentType.HEAT_PUMP_BUILDING,
        ]:
            # if surplus electricity is available, a part of the component's consumption can be covered onsite
            if available_surplus_electricity_in_watt > 0:
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt - electricity_demand_from_current_input_component_in_watt
                )
                stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            # otherwise all of the component's consumption is taken from grid
            else:
                stsv.set_output_value(
                    output=current_output, value=-electricity_demand_from_current_input_component_in_watt
                )
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt - electricity_demand_from_current_input_component_in_watt
                )

        # these are electricity PRODUCERS
        elif current_component_type == lt.ComponentType.CHP:
            available_surplus_electricity_in_watt = (
                available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
            )
            stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)

        return available_surplus_electricity_in_watt

    def modify_set_temperatures_for_building_components_in_case_of_surplus_electricity_iterativ(
        self,
        available_surplus_electricity_in_watt: float,
        stsv: cp.SingleTimeStepValues,
        inputs_sorted: List[ComponentInput],
        component_types_sorted: List[lt.ComponentType],
    ) -> None:
        """In case surplus electricity is available, modify set temperatures for space heating and domestic hot water heat pumps.

        Like this, the heat pumps will start heating up the water storages and the surplus energy can be stored as thermal energy.
        See also SG-ready heatpumps: https://de.gridx.ai/wissen/sg-ready.

        The temperature modification outputs go to the heat pumps, the heat distribution system and the building component (see network charts).
        """
        for index in range(len(inputs_sorted)):
            current_component_type = component_types_sorted[index]

            if current_component_type == lt.ComponentType.HEAT_PUMP_BUILDING:
                if available_surplus_electricity_in_watt > 0:
                    stsv.set_output_value(
                        self.building_indoor_temperature_modifier,
                        self.building_indoor_temperature_offset_value,
                    )
                    stsv.set_output_value(
                        self.space_heating_water_storage_temperature_modifier,
                        self.space_heating_water_storage_temperature_offset_value,
                    )
                else:
                    stsv.set_output_value(self.building_indoor_temperature_modifier, 0)
                    stsv.set_output_value(self.space_heating_water_storage_temperature_modifier, 0)

            elif current_component_type in [
                lt.ComponentType.HEAT_PUMP_DHW,
                lt.ComponentType.HEAT_PUMP,
            ]:
                if available_surplus_electricity_in_watt > 0:
                    stsv.set_output_value(
                        self.domestic_hot_water_storage_temperature_modifier,
                        self.domestic_hot_water_storage_temperature_offset_value,
                    )
                else:
                    stsv.set_output_value(self.domestic_hot_water_storage_temperature_modifier, 0)

    def distribute_available_surplus_electricity_to_building_components_parallel(
        self,
        components_parameters: dict,
        component_and_number_of_same_source_weights: dict,
        available_surplus_electricity_in_watt: float,
        stsv: cp.SingleTimeStepValues,
    ) -> float:
        """Evaluates available surplus electricity component by component.

        - First step: actual electricity demand of each component is subtracted from surplus
        - second step: available surplus is distributed to components iterative if same source weight

        Parallel distribution of surplus when components have equal source weight.
        To do this, the surplus is divided by the number of components with equal source weight so that each component
        has the chance to receive an equal share. If a component requires less, the rest is returned to available surplus.
        """

        component_electricity_demand: dict = {}

        for element in components_parameters:
            current_electricity_demand = stsv.get_input_value(component_input=components_parameters[element]["input"])
            if components_parameters[element]["component_type"] == "Battery":
                component_electricity_demand[element] = float(current_electricity_demand)
            else:
                component_electricity_demand[element] = -1 * float(abs(current_electricity_demand))

        surplus_next_iteration = 0.0
        additional_electricity_demand = 0.0

        for single_source_weight_sorted in list(component_and_number_of_same_source_weights.keys()):

            available_surplus_electricity_in_watt_next_component = (
                available_surplus_electricity_in_watt + additional_electricity_demand + surplus_next_iteration
            )

            components_with_current_source_weight = component_and_number_of_same_source_weights[
                single_source_weight_sorted
            ]["components"]
            number_of_components_with_current_source_weight = component_and_number_of_same_source_weights[
                single_source_weight_sorted
            ]["count"]

            if number_of_components_with_current_source_weight > 1:

                repeat_loop = True
                repeat_count = 0

                while repeat_loop:
                    repeat_loop = False
                    surplus_next_iteration = 0.0
                    additional_electricity_demand = 0.0
                    counter_inner_interation = 0

                    number_of_components_with_electricity_demand_and_same_source_weight = sum(
                        1
                        for key, wert in component_electricity_demand.items()
                        if wert < 0 and key in components_with_current_source_weight
                    )

                    for single_component in components_with_current_source_weight:
                        single_output_sorted = components_parameters[single_component]["output"]

                        electricity_demand_from_current_input_component_in_watt = component_electricity_demand[
                            single_component
                        ]

                        if repeat_count > 0 and component_electricity_demand[single_component] >= 0:
                            continue

                        if counter_inner_interation == 0:
                            if number_of_components_with_electricity_demand_and_same_source_weight == 0:
                                available_surplus_electricity_in_watt_split = (
                                    available_surplus_electricity_in_watt_next_component
                                    / number_of_components_with_current_source_weight
                                )
                                available_surplus_electricity_in_watt_next_component = (
                                    available_surplus_electricity_in_watt_split
                                )
                            elif number_of_components_with_electricity_demand_and_same_source_weight == 1:
                                available_surplus_electricity_in_watt_split = (
                                    available_surplus_electricity_in_watt_next_component
                                )
                                available_surplus_electricity_in_watt_next_component = 0
                            else:
                                available_surplus_electricity_in_watt_split = (
                                    available_surplus_electricity_in_watt_next_component
                                    / number_of_components_with_electricity_demand_and_same_source_weight
                                )
                                available_surplus_electricity_in_watt_next_component = (
                                    available_surplus_electricity_in_watt_split
                                )
                        else:
                            available_surplus_electricity_in_watt_split = (
                                available_surplus_electricity_in_watt_next_component
                            )

                        available_surplus_electricity_in_watt = self.control_electricity_and_modify_set_temperatures_for_component_parallel(
                            available_surplus_electricity_in_watt=available_surplus_electricity_in_watt_split,
                            stsv=stsv,
                            current_component_type=components_parameters[single_component]["component_type"],
                            current_output=single_output_sorted,
                            electricity_demand_from_current_input_component_in_watt=electricity_demand_from_current_input_component_in_watt,
                        )

                        component_electricity_demand[single_component] = available_surplus_electricity_in_watt

                        if available_surplus_electricity_in_watt > 0:
                            surplus_next_iteration += available_surplus_electricity_in_watt
                            available_surplus_electricity_in_watt = 0
                        else:
                            additional_electricity_demand += available_surplus_electricity_in_watt
                            available_surplus_electricity_in_watt = 0

                        counter_inner_interation += 1

                    if (
                        surplus_next_iteration > 0 > additional_electricity_demand
                        and counter_inner_interation
                        >= number_of_components_with_electricity_demand_and_same_source_weight
                    ):
                        available_surplus_electricity_in_watt_next_component = surplus_next_iteration
                        repeat_count += 1
                        repeat_loop = True

            else:
                surplus_next_iteration = 0.0
                additional_electricity_demand = 0.0
                for single_component in components_with_current_source_weight:

                    single_output_sorted = components_parameters[single_component]["output"]

                    electricity_demand_from_current_input_component_in_watt = component_electricity_demand[
                        single_component
                    ]

                    available_surplus_electricity_in_watt = self.control_electricity_and_modify_set_temperatures_for_component_parallel(
                        available_surplus_electricity_in_watt=available_surplus_electricity_in_watt_next_component,
                        stsv=stsv,
                        current_component_type=components_parameters[single_component]["component_type"],
                        current_output=single_output_sorted,
                        electricity_demand_from_current_input_component_in_watt=electricity_demand_from_current_input_component_in_watt,
                    )

                    component_electricity_demand[single_component] = available_surplus_electricity_in_watt

                    if available_surplus_electricity_in_watt > 0:
                        surplus_next_iteration += available_surplus_electricity_in_watt
                        available_surplus_electricity_in_watt = 0
                    else:
                        additional_electricity_demand += available_surplus_electricity_in_watt
                        available_surplus_electricity_in_watt = 0

        available_surplus_electricity_in_watt = additional_electricity_demand + surplus_next_iteration

        return available_surplus_electricity_in_watt

    def control_electricity_parallel(
        self,
        available_surplus_electricity_in_watt: float,
        stsv: cp.SingleTimeStepValues,
        current_component_type: lt.ComponentType,
        current_output: cp.ComponentOutput,
        electricity_demand_from_current_input_component_in_watt: float,
    ) -> float:
        """Calculates available surplus electricity.

        Subtracts the electricity consumption signal of the component from the previous iteration,
        and sends updated signal back.
        This function controls how surplus electricity is distributed and how much of each components'
        electricity need is covered onsite or from grid.
        Electricity input of each component is saved in dict to make iteration possible when parallel distribution is used.
        """

        # if available_surplus_electricity > 0: electricity is fed into battery
        # if available_surplus_electricity < 0: electricity is taken from battery
        if current_component_type == lt.ComponentType.BATTERY:
            stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            # difference between what is fed into battery and what battery really used
            available_surplus_electricity_in_watt = (
                available_surplus_electricity_in_watt - electricity_demand_from_current_input_component_in_watt
            )

        # Building Tag, if district controller manage the hp in building --> contracting
        # Building has own ems which controll the temperatures of storages
        # District EMS controll how surplus is devided to buildings(ems)
        elif current_component_type == lt.ComponentType.SURPLUS_CONTROLLER_DISTRICT:
            if available_surplus_electricity_in_watt > 0:
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )
                stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            else:
                stsv.set_output_value(
                    output=current_output, value=electricity_demand_from_current_input_component_in_watt
                )
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )

        # these are electricity CONSUMERS
        elif current_component_type in [
            lt.ComponentType.RESIDENTS,
            lt.ComponentType.BUILDINGS,
            lt.ComponentType.ELECTROLYZER,
            lt.ComponentType.SMART_DEVICE,
            lt.ComponentType.CAR_BATTERY,
            lt.ComponentType.HEAT_PUMP_DHW,
            lt.ComponentType.HEAT_PUMP,
            lt.ComponentType.HEAT_PUMP_BUILDING,
        ]:

            # if surplus electricity is available, a part of the component's consumption can be covered onsite
            if available_surplus_electricity_in_watt > 0:
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )
                stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            # otherwise all of the component's consumption is taken from grid
            else:
                stsv.set_output_value(
                    output=current_output, value=electricity_demand_from_current_input_component_in_watt
                )
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )

        # these are electricity PRODUCERS
        elif current_component_type == lt.ComponentType.CHP:
            available_surplus_electricity_in_watt = (
                available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
            )
            stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)

        return available_surplus_electricity_in_watt

    def control_electricity_and_modify_set_temperatures_for_component_parallel(
        self,
        available_surplus_electricity_in_watt: float,
        stsv: cp.SingleTimeStepValues,
        current_component_type: lt.ComponentType,
        current_output: cp.ComponentOutput,
        electricity_demand_from_current_input_component_in_watt: float,
    ) -> float:
        """Calculates available surplus electricity.

        Subtracts the electricity consumption signal of the component from the previous iteration,
        and sends updated signal back.
        This function controls how surplus electricity is distributed and how much of each components'
        electricity need is covered onsite or from grid.
        Electricity input if each component is saved in dict to make iteration possible when parallel distribution is used.
        """

        # if available_surplus_electricity > 0: electricity is fed into battery
        # if available_surplus_electricity < 0: electricity is taken from battery
        if current_component_type == lt.ComponentType.BATTERY:
            stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            # difference between what is fed into battery and what battery really used
            available_surplus_electricity_in_watt = (
                available_surplus_electricity_in_watt - electricity_demand_from_current_input_component_in_watt
            )

        # Building Tag, if district controller manage the hp in building --> contracting
        # Building has own ems which controll the temperatures of storages
        # District EMS controll how surplus is devided to buildings(ems)
        elif current_component_type == lt.ComponentType.SURPLUS_CONTROLLER_DISTRICT:
            if available_surplus_electricity_in_watt > 0:
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )
                stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            else:
                stsv.set_output_value(
                    output=current_output, value=electricity_demand_from_current_input_component_in_watt
                )
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )

        # these are electricity CONSUMERS
        elif current_component_type in [
            lt.ComponentType.RESIDENTS,
            lt.ComponentType.BUILDINGS,
            lt.ComponentType.ELECTROLYZER,
            lt.ComponentType.SMART_DEVICE,
            lt.ComponentType.CAR_BATTERY,
            lt.ComponentType.HEAT_PUMP_DHW,
            lt.ComponentType.HEAT_PUMP,
            lt.ComponentType.HEAT_PUMP_BUILDING,
        ]:

            # if surplus electricity is available, a part of the component's consumption can be covered onsite
            if available_surplus_electricity_in_watt > 0:
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )
                stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)
            # otherwise all of the component's consumption is taken from grid
            else:
                stsv.set_output_value(
                    output=current_output, value=electricity_demand_from_current_input_component_in_watt
                )
                available_surplus_electricity_in_watt = (
                    available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
                )

        # these are electricity PRODUCERS
        elif current_component_type == lt.ComponentType.CHP:
            available_surplus_electricity_in_watt = (
                available_surplus_electricity_in_watt + electricity_demand_from_current_input_component_in_watt
            )
            stsv.set_output_value(output=current_output, value=available_surplus_electricity_in_watt)

        # modify set temperatures directly to use surplus in right order of source weights
        if current_component_type == lt.ComponentType.HEAT_PUMP_BUILDING:
            if available_surplus_electricity_in_watt > 0:
                stsv.set_output_value(
                    self.building_indoor_temperature_modifier,
                    self.building_indoor_temperature_offset_value,
                )
                stsv.set_output_value(
                    self.space_heating_water_storage_temperature_modifier,
                    self.space_heating_water_storage_temperature_offset_value,
                )
            else:
                stsv.set_output_value(self.building_indoor_temperature_modifier, 0)
                stsv.set_output_value(self.space_heating_water_storage_temperature_modifier, 0)

        elif current_component_type in [
            lt.ComponentType.HEAT_PUMP_DHW,
            lt.ComponentType.HEAT_PUMP,
        ]:
            if available_surplus_electricity_in_watt > 0:
                stsv.set_output_value(
                    self.domestic_hot_water_storage_temperature_modifier,
                    self.domestic_hot_water_storage_temperature_offset_value,
                )
            else:
                stsv.set_output_value(self.domestic_hot_water_storage_temperature_modifier, 0)

        return available_surplus_electricity_in_watt

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""

        advanced_heat_pump_class_name = advanced_heat_pump_hplib.HeatPumpHplib.get_classname()
        more_advanced_heat_pump_class_name = more_advanced_heat_pump_hplib.MoreAdvancedHeatPumpHPLib.get_classname()
        dhw_heat_pump_class_name = generic_heat_pump_modular.ModularHeatPump.get_classname()
        occupancy_class_name = loadprofilegenerator_utsp_connector.UtspLpgConnector.get_classname()

        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                if dhw_heat_pump_class_name in output.field_name:
                    dhw_hp_electricity_from_grid_in_watt_series = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] < 0.0
                    ]
                    dhw_heatpump_electricity_from_grid_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=dhw_hp_electricity_from_grid_in_watt_series,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    dhw_heatpump_electricity_from_grid_entry = KpiEntry(
                        name=f"Domestic hot water heat pump electricity from grid {self.component_name}",
                        unit="kWh",
                        value=dhw_heatpump_electricity_from_grid_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(dhw_heatpump_electricity_from_grid_entry)
                elif more_advanced_heat_pump_class_name in output.field_name:
                    if "SH" in str(output.field_name):
                        sh_electricity_from_grid_in_watt_series = postprocessing_results.iloc[:, index].loc[
                            postprocessing_results.iloc[:, index] < 0.0
                        ]
                        sh_heatpump_electricity_from_grid_in_kilowatt_hour = abs(
                            KpiHelperClass.compute_total_energy_from_power_timeseries(
                                power_timeseries_in_watt=sh_electricity_from_grid_in_watt_series,
                                timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                            )
                        )
                        # make kpi entry
                        sh_heatpump_electricity_from_grid_entry = KpiEntry(
                            name=f"Space heating heat pump electricity from grid {self.component_name}",
                            unit="kWh",
                            value=sh_heatpump_electricity_from_grid_in_kilowatt_hour,
                            tag=KpiTagEnumClass.EMS,
                            description=self.component_name,
                        )
                        list_of_kpi_entries.append(sh_heatpump_electricity_from_grid_entry)
                    elif "DHW" in output.field_name:
                        dhw_hp_electricity_from_grid_in_watt_series = postprocessing_results.iloc[:, index].loc[
                            postprocessing_results.iloc[:, index] < 0.0
                        ]
                        dhw_heatpump_electricity_from_grid_in_kilowatt_hour = abs(
                            KpiHelperClass.compute_total_energy_from_power_timeseries(
                                power_timeseries_in_watt=dhw_hp_electricity_from_grid_in_watt_series,
                                timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                            )
                        )
                        dhw_heatpump_electricity_from_grid_entry = KpiEntry(
                            name=f"Domestic hot water heat pump electricity from grid {self.component_name}",
                            unit="kWh",
                            value=dhw_heatpump_electricity_from_grid_in_kilowatt_hour,
                            tag=KpiTagEnumClass.EMS,
                            description=self.component_name,
                        )
                        list_of_kpi_entries.append(dhw_heatpump_electricity_from_grid_entry)
                    else:
                        log.warning(f"No DHW oder SH named in output {output.field_name} of {output.component_name}")
                elif advanced_heat_pump_class_name in output.field_name:
                    sh_electricity_from_grid_in_watt_series = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] < 0.0
                    ]
                    sh_heatpump_electricity_from_grid_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=sh_electricity_from_grid_in_watt_series,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    # make kpi entry
                    sh_heatpump_electricity_from_grid_entry = KpiEntry(
                        name=f"Space heating heat pump electricity from grid {self.component_name}",
                        unit="kWh",
                        value=sh_heatpump_electricity_from_grid_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(sh_heatpump_electricity_from_grid_entry)
                elif occupancy_class_name in output.field_name:
                    occupancy_electricity_from_grid_in_watt_series = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] < 0.0
                    ]

                    occupancy_electricity_from_grid_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=occupancy_electricity_from_grid_in_watt_series,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    occupancy_electricity_from_grid_entry = KpiEntry(
                        name=f"Residents' electricity consumption from grid {self.component_name}",
                        unit="kWh",
                        value=occupancy_electricity_from_grid_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(occupancy_electricity_from_grid_entry)
                elif output.field_name == self.TotalElectricityConsumptionEMSControlled:
                    total_electricity_consumption_ems_controlled_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_electricity_consumption_ems_controlled_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=total_electricity_consumption_ems_controlled_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    electricity_consumption_ems_controlled_entry = KpiEntry(
                        name=f"Controlled electricity consumption {self.component_name}",
                        unit="kWh",
                        value=total_electricity_consumption_ems_controlled_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(electricity_consumption_ems_controlled_entry)
                elif output.field_name == self.TotalElectricityConsumptionUnControlled:
                    total_electricity_consumption_ems_uncontrolled_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_electricity_consumption_ems_uncontrolled_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=total_electricity_consumption_ems_uncontrolled_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    electricity_consumption_ems_uncontrolled_entry = KpiEntry(
                        name=f"Uncontrolled electricity consumption {self.component_name}",
                        unit="kWh",
                        value=total_electricity_consumption_ems_uncontrolled_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(electricity_consumption_ems_uncontrolled_entry)
                elif output.field_name == self.TotalElectricityConsumption:
                    total_electricity_consumption_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_electricity_consumption_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=total_electricity_consumption_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    electricity_consumption_total_entry = KpiEntry(
                        name=f"Total electricity consumption {self.component_name}",
                        unit="kWh",
                        value=total_electricity_consumption_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(electricity_consumption_total_entry)
                elif output.field_name == self.SurplusUnusedFromDistrictEMSOutput:
                    total_electricity_surplus_from_district_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_electricity_surplus_from_district_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=total_electricity_surplus_from_district_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    electricity_surplus_from_district_entry = KpiEntry(
                        name=f"Surplus from District {self.component_name}",
                        unit="kWh",
                        value=total_electricity_surplus_from_district_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(electricity_surplus_from_district_entry)
                elif output.field_name == self.ElectricityProduction:
                    total_production_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_production_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=total_production_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    electricity_production_entry = KpiEntry(
                        name=f"Production {self.component_name}",
                        unit="kWh",
                        value=total_production_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(electricity_production_entry)
                elif output.field_name == self.TotalElectricityFromGrid:
                    total_electricity_from_grid_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_electricity_from_grid_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=total_electricity_from_grid_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    electricity_from_grid_entry = KpiEntry(
                        name=f"Total electricity from grid {self.component_name}",
                        unit="kWh",
                        value=total_electricity_from_grid_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(electricity_from_grid_entry)
                elif output.field_name == self.TotalElectricityToGrid:
                    total_electricity_to_grid_in_watt = postprocessing_results.iloc[:, index].loc[
                        postprocessing_results.iloc[:, index] > 0.0
                    ]
                    total_electricity_to_grid_in_kilowatt_hour = abs(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=total_electricity_to_grid_in_watt,
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        )
                    )
                    electricity_to_grid_entry = KpiEntry(
                        name=f"Total electricity to grid {self.component_name}",
                        unit="kWh",
                        value=total_electricity_to_grid_in_kilowatt_hour,
                        tag=KpiTagEnumClass.EMS,
                        description=self.component_name,
                    )
                    list_of_kpi_entries.append(electricity_to_grid_entry)

        return list_of_kpi_entries

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> cp.OpexCostDataClass:
        """Calculate OPEX costs, consisting of electricity costs and revenues."""
        opex_cost_data_class = cp.OpexCostDataClass.get_default_opex_cost_data_class()
        return opex_cost_data_class

    @staticmethod
    def get_cost_capex(config: EMSDistrictConfig, simulation_parameters: SimulationParameters) -> cp.CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = cp.CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class


"""
                ***IN PROGRESS***
    def distribute_available_surplus_from_district_to_building_parallel(
        self,
        source_weights_sorted: List[int],
        available_surplus_electricity_in_watt: float,
        stsv: cp.SingleTimeStepValues,
        inputs_sorted: List[ComponentInput],
        component_types_sorted: List[lt.ComponentType],
        outputs_sorted: List[ComponentOutput],
    ) -> float:
        """ """Evaluates available surplus electricity component by component.

        - First step: actual electricity demand of each component is subtracted from surplus
        - second step: available surplus is distributed to components iterative if same source weight

        Parallel distribution of surplus when components have equal source weight.
        To do this, the surplus is divided by the number of components with equal source weight so that each component
        has the chance to receive an equal share. If a component requires less, the rest is returned to available surplus.
        """ """
        component_types_and_number_of_same_source_weights: dict = {}

        for weight, component in zip(source_weights_sorted, component_types_sorted):
            if weight in component_types_and_number_of_same_source_weights:
                component_types_and_number_of_same_source_weights[weight]["count"] += 1
                component_types_and_number_of_same_source_weights[weight]["components"].append(component)
            else:
                component_types_and_number_of_same_source_weights[weight] = {"count": 1, "components": [component]}

        component_electricity_demand: dict = {}

        for index, item in enumerate(component_types_sorted):
            value = stsv.get_input_value(component_input=inputs_sorted[index])
            component_electricity_demand[item] = -1 * float(abs(value))

        repeat_count = 0
        previous_single_source_weight_sorted = 0

        available_surplus_electricity_in_watt_next_component = available_surplus_electricity_in_watt

        # if (available_surplus_electricity_in_watt > 100 and list(component_electricity_demand.values())[
        #          0] == 0 and list(component_electricity_demand.values())[1] < -100) :
        #     print("debug")

        repeat_max = max(
            component_types_and_number_of_same_source_weights[single_source_weight_sorted]["count"]
            for single_source_weight_sorted in source_weights_sorted
        )

        while repeat_count < repeat_max + 5:
            repeat_loop = False
            index = 0
            surplus_next_iteration = 0.0
            additional_electricity_demand = 0.0
            counter_inner_while = 0

            while index < len(inputs_sorted):
                single_component_type_sorted = component_types_sorted[index]
                single_output_sorted = outputs_sorted[index]
                single_source_weight_sorted = source_weights_sorted[index]

                number_of_components_with_same_source_weights = component_types_and_number_of_same_source_weights[
                    single_source_weight_sorted
                ]["count"]

                electricity_demand_from_current_input_component_in_watt = component_electricity_demand[
                    single_component_type_sorted
                ]

                index += 1

                if single_source_weight_sorted > previous_single_source_weight_sorted:
                    counter_inner_while = 0

                available_surplus_electricity_in_watt = available_surplus_electricity_in_watt_next_component

                #
                #  if repeat_count > 0:
                #      print("debug")

                # if (
                #     component_electricity_demand[single_component_type_sorted] >= 0
                #     or number_of_components_with_electricity_demand_and_same_source_weight == 0
                # ):
                #     available_surplus_electricity_in_watt_next_component = available_surplus_electricity_in_watt
                #     additional_electricity_demand = additional_electricity_demand
                #     surplus_next_iteration = surplus_next_iteration
                #     continue

                if repeat_count > 0 and component_electricity_demand[single_component_type_sorted] >= 0:
                    available_surplus_electricity_in_watt_next_component = available_surplus_electricity_in_watt
                    additional_electricity_demand = additional_electricity_demand
                    surplus_next_iteration = surplus_next_iteration
                    component_electricity_demand[single_component_type_sorted] = 0
                    continue

                if component_types_and_number_of_same_source_weights[single_source_weight_sorted]["count"] > 1:

                    number_of_components_with_electricity_demand_and_same_source_weight = sum(
                        1
                        for key, wert in component_electricity_demand.items()
                        if wert < 0
                        and key
                        in component_types_and_number_of_same_source_weights[single_source_weight_sorted]["components"]
                    )

                    if counter_inner_while == 0:

                        if number_of_components_with_electricity_demand_and_same_source_weight <= 1:
                            available_surplus_electricity_in_watt_split = available_surplus_electricity_in_watt
                            available_surplus_electricity_in_watt_next_component = 0
                        # if number_of_components_with_electricity_demand_and_same_source_weight <= 1:
                        #     available_surplus_electricity_in_watt_split = available_surplus_electricity_in_watt
                        #     available_surplus_electricity_in_watt_next_component = 0
                        else:
                            available_surplus_electricity_in_watt_split = (
                                available_surplus_electricity_in_watt
                                / number_of_components_with_electricity_demand_and_same_source_weight
                            )
                            # if number_of_components_with_electricity_demand_and_same_source_weight == 1:
                            #     available_surplus_electricity_in_watt_next_component = 0
                            # else:
                            available_surplus_electricity_in_watt_next_component = (
                                available_surplus_electricity_in_watt_split
                            )

                    else:
                        available_surplus_electricity_in_watt_split = (
                            available_surplus_electricity_in_watt_next_component
                        )

                    available_surplus_electricity_in_watt = self.control_electricity_parallel(
                        available_surplus_electricity_in_watt=available_surplus_electricity_in_watt_split,
                        stsv=stsv,
                        current_component_type=single_component_type_sorted,
                        current_output=single_output_sorted,
                        component_electricity_demand=component_electricity_demand,
                        electricity_demand_from_current_input_component_in_watt=electricity_demand_from_current_input_component_in_watt,
                    )

                    counter_inner_while += 1

                    if available_surplus_electricity_in_watt > 0:
                        surplus_next_iteration += available_surplus_electricity_in_watt
                        available_surplus_electricity_in_watt = 0
                    else:
                        additional_electricity_demand += available_surplus_electricity_in_watt
                        available_surplus_electricity_in_watt = 0

                    if (
                        surplus_next_iteration > 0
                        and additional_electricity_demand < 0
                        and counter_inner_while == number_of_components_with_same_source_weights
                    ):
                        repeat_loop = True

                else:

                    available_surplus_electricity_in_watt = (
                        available_surplus_electricity_in_watt + surplus_next_iteration + additional_electricity_demand
                    )

                    available_surplus_electricity_in_watt = self.control_electricity_parallel(
                        available_surplus_electricity_in_watt=available_surplus_electricity_in_watt,
                        stsv=stsv,
                        current_component_type=single_component_type_sorted,
                        current_output=single_output_sorted,
                        component_electricity_demand=component_electricity_demand,
                        electricity_demand_from_current_input_component_in_watt=electricity_demand_from_current_input_component_in_watt,
                    )

                    available_surplus_electricity_in_watt_next_component = available_surplus_electricity_in_watt

                previous_single_source_weight_sorted = single_source_weight_sorted

                if repeat_loop:
                    break

            if repeat_loop:
                repeat_count += 1
                available_surplus_electricity_in_watt_next_component = surplus_next_iteration

            else:
                break

        # available_surplus_electricity_in_watt = sum(
        #     values for values in component_electricity_demand.values()
        # )  # if values < 0)

        max_key = max(component_types_and_number_of_same_source_weights.keys())
        components_with_max_count = component_types_and_number_of_same_source_weights[max_key]["components"]

        available_surplus_electricity_in_watt = sum(
            component_electricity_demand[component] for component in components_with_max_count
        )

        return available_surplus_electricity_in_watt
"""



"""Air Conditioner Component."""

# clean

from dataclasses import dataclass
from typing import Any
from dataclasses_json import dataclass_json

# from typing import Optional
import numpy as np
from hisim import log

# owned
from hisim import component as cp
from hisim.component import ConfigBase, DisplayConfig
from hisim.simulationparameters import SimulationParameters
from hisim.loadtypes import LoadTypes, Units
from hisim.components.weather import Weather
from hisim.components.building import Building
from hisim.components.controller_pid import PIDController
from hisim import utils
from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum

__authors__ = "Marwa Alfouly"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Marwa Alfouly"
__email__ = "m.alfouly@fz-juelich.de"
__status__ = "development"


@dataclass_json
@dataclass
class AirConditionerConfig(ConfigBase):
    """Class for configuration of air-conditioner."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return AirConditioner.get_full_classname()

    building_name: str
    name: str
    manufacturer: str
    model_name: str
    min_operation_time: int
    min_idle_time: int
    control: str
    # my_simulation_repository: Optional[cp.SimRepository] = None

    @classmethod
    def get_default_air_conditioner_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Get default configuration of air-conditioner."""
        config = AirConditionerConfig(
            building_name=building_name,
            name="AirConditioner",
            manufacturer="Panasonic",
            model_name="CS-RE18JKE/CU-RE18JKE",
            min_operation_time=60 * 60,
            min_idle_time=15 * 60,
            control="on_off",
            # my_simulation_repository=None,
        )
        return config


@dataclass_json
@dataclass
class AirConditionerControllerConfig(ConfigBase):
    """Class for configuration of air-conditioner controller."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return AirConditionerController.get_full_classname()

    building_name: str
    name: str
    t_air_heating: float
    t_air_cooling: float
    offset: float

    @classmethod
    def get_default_air_conditioner_controller_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Get default configuration of air-conditioner controller."""
        config = AirConditionerControllerConfig(
            building_name=building_name,
            name="AirConditioner",
            t_air_heating=18.0,
            t_air_cooling=26.0,
            offset=0.0,
        )
        return config


class AirConditionerState:
    """Data class for saving the state of the air conditioner."""

    def __init__(
        self,
        timestep_actual: int = -1,
        state: int = 0,
        timestep_of_last_action: int = 0,
    ):
        """Initializes the state of the air conditioner."""
        self.timestep_actual = timestep_actual
        self.state = state
        self.timestep_of_last_action = timestep_of_last_action

    def clone(self):
        """Storing last timestep state."""
        return AirConditionerState(
            timestep_actual=self.timestep_actual,
            state=self.state,
            timestep_of_last_action=self.timestep_of_last_action,
        )

    def is_first_iteration(self, timestep):
        """Check if it is the first iteration in order to ensure minimal on and off times."""
        bool_is_first_iteration = False
        if self.timestep_actual + 1 == timestep:
            self.timestep_actual += 1
            bool_is_first_iteration = True
        else:
            bool_is_first_iteration = False
        return bool_is_first_iteration

    def activation(self, timestep):
        """Take action based on the operation time / idle time."""
        self.state = 1
        self.timestep_of_last_action = timestep

    def deactivation(self, timestep):
        """Take action based on the operation time / idle time."""
        self.state = 0
        self.timestep_of_last_action = timestep


class AirConditioner(cp.Component):
    """Class for air-conditioner."""

    # inputs
    State = "State"
    # weather
    TemperatureOutside = "TemperatureOutside"
    # controller_pid
    ThermalPowerPID = "ThermalPowerPID"
    # controller_mpc
    OperatingMode = "OperatingMode"
    GridImport = "GridImport"
    PV2load = "PV2load"
    Battery2Load = "Battery2Load"
    # building
    TemperatureMean = "Residence Temperature"
    FeedForwardSignal = "FeedForwardSignal"
    # ElectricityOutputPID = "ElectricityOutputPID"

    # outputs
    ThermalEnergyDelivered = "ThermalEnergyDelivered"
    ElectricityOutput = "ElectricityOutput"
    PidManipulatedVariable = "PidManipulatedVariable"
    EER = "EER"

    cop_coef_heating = "cop_coef_heating"
    eer_coef_cooling = "eer_coef_cooling"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: AirConditionerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Constructs all the necessary attributes."""

        self.air_conditioner_config = config
        # self.state = 0
        # self.state0 = 0
        # self.previous_state = 0
        self.state = AirConditionerState()
        self.state0 = AirConditionerState()
        self.previous_state = AirConditionerState()

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.build(
            manufacturer=self.air_conditioner_config.manufacturer,
            model_name=self.air_conditioner_config.model_name,
            min_operation_time=self.air_conditioner_config.min_operation_time,
            min_idle_time=self.air_conditioner_config.min_idle_time,
            # my_simulation_repository=self.air_conditioner_config.my_simulation_repository,
        )
        self.t_out_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureOutside,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            True,
        )
        self.t_m_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureMean,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            True,
        )
        self.state_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.State,
            LoadTypes.ANY,
            Units.ANY,
            False,
        )
        self.feed_forward_signal_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.FeedForwardSignal,
            LoadTypes.HEATING,
            Units.WATT,
            False,
        )
        self.thermal_power: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ThermalPowerPID,
            LoadTypes.HEATING,
            Units.WATT,
            False,
        )
        self.operating_mode_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.OperatingMode,
            LoadTypes.ANY,
            Units.ANY,
            False,
        )
        self.optimal_electric_power_pv_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.PV2load,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            False,
        )
        self.optimal_electric_power_grid_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.GridImport,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            False,
        )
        self.optimal_electric_power_battery_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.Battery2Load,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            False,
        )
        self.pid_thermal_power_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.PidManipulatedVariable,
            LoadTypes.HEATING,
            Units.WATT,
            output_description=f"here a description for {self.PidManipulatedVariable} will follow.",
        )
        self.thermal_energy_delivered_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ThermalEnergyDelivered,
            LoadTypes.HEATING,
            Units.WATT,
            output_description=f"here a description for Air Conditioner {self.ThermalEnergyDelivered} will follow.",
        )
        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ElectricityOutput,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for Air Conditioner {self.ElectricityOutput} will follow.",
        )
        self.cooling_eer_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.EER,
            LoadTypes.ANY,
            Units.ANY,
            output_description=f"here a description for {self.EER} will follow.",
        )

        self.add_default_connections(self.get_default_connections_from_weather())
        self.add_default_connections(self.get_default_connections_from_air_condition_controller())

        self.control = self.air_conditioner_config.control

    def get_default_connections_from_weather(self):
        """Get default inputs from the weather component."""

        connections = []
        weather_classname = Weather.get_classname()
        connections.append(
            cp.ComponentConnection(
                AirConditioner.TemperatureOutside,
                weather_classname,
                Weather.TemperatureOutside,
            )
        )
        return connections

    def get_default_connections_from_air_condition_controller(self):
        """Get default inputs from the on_off controller."""

        connections = []
        controller_classname = AirConditionerController.get_classname()
        connections.append(
            cp.ComponentConnection(
                AirConditioner.State,
                controller_classname,
                AirConditionerController.State,
            )
        )
        return connections

    def get_pid_controller_default_connections(self):
        """Get default inputs from the PID controller component."""

        connections = []
        controller_classname = AirConditionerController.get_classname()
        connections.append(
            cp.ComponentConnection(
                AirConditioner.ThermalPowerPID,
                controller_classname,
                PIDController.ThermalPowerPID,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def build(
        self,
        manufacturer,
        model_name,
        min_operation_time,
        min_idle_time,
        # my_simulation_repository,
    ):
        """Build function: The function retrieves air conditioner from databasesets sets important constants and parameters for the calculations."""
        # Simulation parameters

        # Retrieves air conditioner from database - BEGIN
        air_conditioners_database = utils.load_smart_appliance("Air Conditioner")

        air_conditioner = None
        for air_conditioner_iterator in air_conditioners_database:
            if (
                air_conditioner_iterator["Manufacturer"] == manufacturer
                and air_conditioner_iterator["Model"] == model_name
            ):
                air_conditioner = air_conditioner_iterator
                break

        if air_conditioner is None:
            raise Exception("Air conditioner model not registered in the database")

        self.manufacturer = manufacturer
        self.model = model_name
        self.min_operation_time = min_operation_time
        self.min_idle_time = min_idle_time

        # Interpolates COP, cooling capacities, power input data from the database
        self.eer_ref = []
        self.t_out_cooling_ref = []
        self.t_out_heating_ref = []
        self.cooling_capacity_ref = []
        self.heating_capacity_ref = []
        self.cop_ref = []

        """
        Typical realtion between COPs and Heating capacities are found here:  https://www.everysolarthing.com/blog/heat-pumps/

        """

        for air_conditioner_tout in air_conditioner["Outdoor temperature range - cooling"]:
            self.t_out_cooling_ref.append([air_conditioner_tout][0])
        for air_conditioner_tout in air_conditioner["Outdoor temperature range - heating"]:
            self.t_out_heating_ref.append([air_conditioner_tout][0])

        for air_conditioner_eers in air_conditioner["EER W/W"]:
            self.eer_ref.append([air_conditioner_eers][0])
        for air_conditioner_cops in air_conditioner["COP W/W"]:
            self.cop_ref.append([air_conditioner_cops][0])

        for air_conditioner_cooling_capacities in air_conditioner["Cooling capacity W"]:
            self.cooling_capacity_ref.append([air_conditioner_cooling_capacities][0])
        for air_conditioner_heating_capacities in air_conditioner["Heating capacity W"]:
            self.heating_capacity_ref.append([air_conditioner_heating_capacities][0])
        print(str(self.t_out_cooling_ref))
        print(str(self.eer_ref))
        self.eer_coef = np.polyfit(self.t_out_cooling_ref, self.eer_ref, 1)
        self.cooling_capacity_coef = np.polyfit(self.t_out_cooling_ref, self.cooling_capacity_ref, 1)

        self.cop_coef = np.polyfit(self.t_out_heating_ref, self.cop_ref, 1)
        self.heating_capacity_coef = np.polyfit(self.t_out_heating_ref, self.heating_capacity_ref, 1)

        # Retrieves air conditioner from database - END

        SingletonSimRepository().set_entry(key=SingletonDictKeyEnum.COPCOEFFICIENTHEATING, entry=self.cop_coef)
        SingletonSimRepository().set_entry(key=SingletonDictKeyEnum.EERCOEFFICIENTCOOLING, entry=self.eer_coef)
        # self.simulation_repository.set_entry(self.cop_coef_heating, self.cop_coef)
        # self.simulation_repository.set_entry(self.eer_coef_cooling, self.eer_coef)

        # Sets the time operation restricitions
        self.on_time = self.min_operation_time / self.my_simulation_parameters.seconds_per_timestep
        self.off_time = self.min_idle_time / self.my_simulation_parameters.seconds_per_timestep

        # self.state = AirConditionerState()
        # self.previous_state = AirConditionerState()

    def cal_eer(self, t_out):
        """Calculate cooling energy efficiency ratio as a function of outside temperature."""
        return np.polyval(self.eer_coef, t_out)

    def cal_cooling_capacity(self, t_out):
        """Calculate cooling capacity as a function of outside temperature."""
        return np.polyval(self.cooling_capacity_coef, t_out)

    def cal_cop(self, t_out):
        """Calculate heating coefficient of performance as a function of outside temperature."""
        return np.polyval(self.cop_coef, t_out)

    def cal_heating_capacity(self, t_out):
        """Calculate heating capacity as a function of outside temperature."""
        return np.polyval(self.heating_capacity_coef, t_out)

    def i_save_state(self) -> None:
        """Saves the internal state at the beginning of each timestep."""
        self.previous_state = self.state.clone()
        pass

    def i_restore_state(self) -> None:
        """Restores the internal state after each iteration."""
        self.state = self.previous_state.clone()
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Double check results after iteration."""
        pass

    def write_to_report(self):
        """Logs the most important config stuff to the report."""
        lines = []
        lines.append("Name: Air Conditioner")
        lines.append(f"Manufacturer: {self.manufacturer}")
        lines.append(f"Model {self.model}")
        lines.append(f"Min Operation Time [Sec]: {self.min_operation_time}")
        lines.append(f"Min Idle Time [Sec]: {self.min_idle_time}")
        lines.append(f"Control: {self.control}")
        return self.air_conditioner_config.get_string_dict() + lines

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:  # noqa: C901
        """Core simulation function."""
        # Inputs
        t_out = stsv.get_input_value(self.t_out_channel)
        on_off_state = stsv.get_input_value(self.state_channel)

        if self.control == "on_off":
            # Heating Season:
            cop = self.cal_cop(t_out)
            heating_power = self.cal_heating_capacity(t_out)

            # Cooling Season:
            eer = self.cal_eer(t_out)
            cooling_power = self.cal_cooling_capacity(t_out)

            # save reference state state0 in first iteration
            if self.state.is_first_iteration(timestep):
                self.state0 = self.state.clone()

            # return device on if minimum operation time is not fulfilled and device was on in previous state
            if self.state0.state == 1 and self.state0.timestep_of_last_action + self.on_time >= timestep:
                self.state.state = 1
            elif self.state0.state == -1 and self.state0.timestep_of_last_action + self.on_time >= timestep:
                self.state.state = -1
            # return device off if minimum idle time is not fulfilled and device was off in previous state
            elif self.state0.state == 0 and self.state0.timestep_of_last_action + self.off_time >= timestep:
                self.state.state = 0
            # check signal from l2 and turn on or off if it is necesary
            else:
                if on_off_state == 0 and (self.state0.state in (1, -1)):
                    self.state.deactivation(timestep)
                elif (on_off_state in (1, -1)) and self.state0.state == 0:
                    self.state.activation(timestep)

            if self.state.state == 1 and on_off_state == 1:
                thermal_energy_delivered = on_off_state * heating_power
                electricity_output = heating_power / cop
            elif self.state.state == 1 and on_off_state == -1:
                thermal_energy_delivered = on_off_state * cooling_power
                electricity_output = cooling_power / eer
            else:
                thermal_energy_delivered = 0
                electricity_output = 0

            # log.information("thermal_energy_delivered {}".format(thermal_energy_delivered))
            stsv.set_output_value(self.thermal_energy_delivered_channel, thermal_energy_delivered)
            stsv.set_output_value(self.electricity_output_channel, electricity_output)
            stsv.set_output_value(self.cooling_eer_channel, eer)

        if self.control == "PID":
            feed_forward_signal = stsv.get_input_value(self.feed_forward_signal_channel)
            thermal_load = stsv.get_input_value(self.thermal_power)

            if thermal_load > 0:
                cop = self.cal_cop(t_out)
                thermal_energy_delivered = thermal_load + feed_forward_signal
                electricity_output = thermal_energy_delivered / cop
            elif thermal_load < 0:
                eer = self.cal_eer(t_out)
                thermal_energy_delivered = thermal_load + feed_forward_signal
                electricity_output = -thermal_energy_delivered / eer
            else:
                thermal_energy_delivered = 0
                electricity_output = 0

            if thermal_energy_delivered > 15000:
                thermal_energy_delivered = 15000
            elif thermal_energy_delivered < -15000:
                thermal_energy_delivered = -15000

            stsv.set_output_value(self.electricity_output_channel, electricity_output)
            stsv.set_output_value(self.thermal_energy_delivered_channel, thermal_energy_delivered)

        if self.control == "MPC":
            mode = stsv.get_input_value(self.operating_mode_channel)
            optimal_electric_power_grid = stsv.get_input_value(self.optimal_electric_power_grid_channel)
            optimal_electric_power_pv = stsv.get_input_value(self.optimal_electric_power_pv_channel)
            optimal_electric_power_battery = stsv.get_input_value(self.optimal_electric_power_battery_channel)
            optimal_electric_power_total = (
                optimal_electric_power_grid + optimal_electric_power_pv + optimal_electric_power_battery
            )

            if mode == 1:
                cop = self.cal_cop(t_out)
                thermal_energy_delivered = optimal_electric_power_total * cop
                electricity_output = optimal_electric_power_total
            elif mode == -1:
                eer = self.cal_eer(t_out)
                thermal_energy_delivered = -optimal_electric_power_total * eer
                electricity_output = optimal_electric_power_total
            else:
                thermal_energy_delivered = 0
                electricity_output = 0

            stsv.set_output_value(self.electricity_output_channel, electricity_output)
            stsv.set_output_value(self.thermal_energy_delivered_channel, thermal_energy_delivered)


class AirConditionerController(cp.Component):
    """Class for air-conditioner controller."""

    """It takes data from other
    components and sends signal to the air conditioner for
    activation or deactivation.

    Parameters
    --------------
    t_air_cooling: float
        Maximum comfortable temperature for residents
    offset: float
        Temperature offset to compensate the hysteresis
        correction for the building temperature change
    mode : int
        Mode index for operation type for this air conditioner
    """

    # Inputs
    TemperatureMean = "Residence Temperature"
    ElectricityInput = "ElectricityInput"

    # Outputs
    State = "State"

    # Similar components to connect to:
    # 1. Building
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: AirConditionerControllerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Constructs all the neccessary attributes."""

        self.air_conditioner_controller_config = config
        self.aircon_controller_mode: str = "not initalized yet"
        self.previous_aircon_controller_mode: str = "not initalized yet"

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.build(
            t_air_cooling=self.air_conditioner_controller_config.t_air_cooling,
            t_air_heating=self.air_conditioner_controller_config.t_air_heating,
            offset=self.air_conditioner_controller_config.offset,
        )

        self.t_m_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureMean,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            True,
        )
        self.electricity_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityInput,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            False,
        )
        self.state_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.State,
            LoadTypes.ANY,
            Units.ANY,
            output_description=f"here a description for {self.State} will follow.",
        )

        self.add_default_connections(self.get_default_connections_from_building())

    def get_default_connections_from_building(self):
        """Get default inputs from the building component."""
        log.information("setting building default connections in AirConditionerController")
        connections = []
        building_classname = Building.get_classname()
        connections.append(
            cp.ComponentConnection(
                AirConditionerController.TemperatureMean,
                building_classname,
                Building.TemperatureMean,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def build(self, t_air_heating, t_air_cooling, offset):
        """Build function: important settings such as comforatable temperature range and deadband width."""
        # Sth
        self.aircon_controller_mode = "off"
        self.previous_aircon_controller_mode = self.aircon_controller_mode

        # Configuration
        self.t_set_heating = t_air_heating
        self.t_set_cooling = t_air_cooling
        self.offset = offset

    def i_save_state(self):
        """Saves the internal state at the beginning of each timestep."""
        self.previous_aircon_controller_mode = self.aircon_controller_mode

    def i_restore_state(self):
        """Restores the internal state after each iteration."""
        self.aircon_controller_mode = self.previous_aircon_controller_mode

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Double check results after iteration."""
        pass

    def write_to_report(self):
        """Logs the most important config stuff to the report."""
        lines = []
        lines.append("Air Conditioner Controller")
        lines.append("Control algorith of the Air conditioner is: on-off control\n")
        lines.append(f"Controller heating set temperature is {format(self.t_set_heating)} Deg C \n")
        lines.append(f"Controller cooling set temperature is {format(self.t_set_cooling)} Deg C \n")
        return self.air_conditioner_controller_config.get_string_dict() + lines

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Core simulation."""
        # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
        if force_convergence:
            pass
        else:
            # Retrieves inputs
            t_m_old = stsv.get_input_value(self.t_m_channel)
            # electricity_input = stsv.get_input_value(self.electricity_inputC)

            self.conditions(t_m_old)

        if self.aircon_controller_mode == "heating":
            state = 1
        if self.aircon_controller_mode == "cooling":
            state = -1
        if self.aircon_controller_mode == "off":
            state = 0

        stsv.set_output_value(self.state_channel, state)
        # log.information("state {}".format(state))

    def conditions(self, set_temp):
        """Controller takes action to maintain defined comfort range given a certain deadband."""
        maximum_heating_set_temp = self.t_set_heating + self.offset
        minimum_heating_set_temp = self.t_set_heating
        minimum_cooling_set_temp = self.t_set_cooling - self.offset
        maximum_cooling_set_temp = self.t_set_cooling

        if self.aircon_controller_mode == "heating":  # and daily_avg_temp < 15:
            if set_temp > maximum_heating_set_temp:  # 16.5
                self.aircon_controller_mode = "off"
                return
        if self.aircon_controller_mode == "cooling":
            if set_temp < minimum_cooling_set_temp:  # 23.5
                self.aircon_controller_mode = "off"
                return
        if self.aircon_controller_mode == "off":
            # if pvs_surplus > ? and air_temp < minimum_heating_air + 2:
            if set_temp < minimum_heating_set_temp:  # 21
                self.aircon_controller_mode = "heating"
                return
            if set_temp > maximum_cooling_set_temp:  # 26
                self.aircon_controller_mode = "cooling"
                return

    def prin1t_outpu1t(self, t_m, state):
        """Print output of air-conditioner controller."""
        log.information("==========================================")
        log.information(f"T m: {format(t_m)}")
        log.information(f"State: {format(state)}")



"""Simple implementation of an electrolyzer.

The electrolyzer can moderate in a certain range,but the efficiency changes only
linarly. Recovery heat is not considered so far.
"""

from dataclasses import dataclass
from typing import List

from dataclasses_json import dataclass_json

from hisim import component as cp
from hisim import loadtypes as lt
from hisim.components import controller_l1_electrolyzer
from hisim.simulationparameters import SimulationParameters

__authors__ = "Frank Burkrad, Maximilian Hillen"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Johanna Ganglbauer"
__email__ = "johanna.ganglbauer@4wardenergy.at"
__status__ = ""


@dataclass_json
@dataclass
class GenericElectrolyzerConfig(cp.ConfigBase):
    """Generic electrolyzer config."""

    building_name: str
    #: name of the electrolyer
    name: str
    #: priority of the component in hierachy: the higher the number the lower the priority
    source_weight: int
    #: minimal operating power in Watt (electrical power)
    min_power: float
    #: maximal operating power in Watt (electrical power)
    max_power: float
    #: minimal hydrogen production rate (at minimal electrical power) in kg / s
    min_hydrogen_production_rate: float
    #: maximal hydrogen production rate (at maximal electrical power) in kg / s
    max_hydrogen_production_rate: float

    @staticmethod
    def get_default_config(
        p_el: float,
        building_name: str = "BUI1",
    ) -> "GenericElectrolyzerConfig":
        """Returns the default configuration of an electrolyzer."""
        config = GenericElectrolyzerConfig(
            building_name=building_name,
            name="Electrolyzer",
            source_weight=1,
            min_power=p_el * 0.5,
            max_power=p_el,
            min_hydrogen_production_rate=p_el * (1 / 4) * 8.989 / 3.6e4,
            max_hydrogen_production_rate=p_el * (50 / 24) * 8.989 / 3.6e4,
        )
        return config


class ElectrolyzerState:
    """Saves the state of the electrolyzer."""

    def __init__(self, hydrogen: float = 0, electricity: float = 0):
        """Initialize an instance."""
        self.hydrogen = hydrogen
        self.electricity = electricity

    def clone(self) -> "ElectrolyzerState":
        """Return a second instance of this class with same attributes."""
        return ElectrolyzerState(hydrogen=self.hydrogen, electricity=self.electricity)


class GenericElectrolyzer(cp.Component):
    """Generic electrolyzer component.

    The electrolyzer converts electrical energy [kWh] into hydrogen [kg]. It can
    work in a certain range from x to 100% or be switched off = 0%. The
    conversion rate is given by the supplier and is directly used. Maybe a
    change to efficiency can be made but its just making things more complex
    with no benefit. Between the given values, the values are calculated by an
    interpolation. If the load curve is linear a fixed factor could be
    calculated. Therefore it has an operational state. All the min values and
    all the max values are connected and the electrolyzer can operate between
    them. The waste energy in electolyzers is not used to provide heat for the
    households demand. Output pressure may be used in the future.

    Components to connect to: (1) Electrolyzer controller (controller_l1_electrolyzer).

    """

    # Inputs
    AvailableElectricity = "AvailbaleElectricity"

    # Outputs
    HydrogenOutput = "HydrogenOutput"
    ElectricityOutput = "ElectricityOutput"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: GenericElectrolyzerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ):
        """Initialize an instance."""

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.config = config
        self.state = ElectrolyzerState()
        self.previous_state = ElectrolyzerState()

        # Intputs
        self.electricity_target_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            GenericElectrolyzer.AvailableElectricity,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )

        # Outputs
        self.hydrogen_output_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            GenericElectrolyzer.HydrogenOutput,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG_PER_SEC,
            output_description="Hydrogen output",
        )
        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=GenericElectrolyzer.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=[
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_EMS_CONTROLLED,
                lt.ComponentType.ELECTROLYZER,
            ],
            output_description="Electricity Output",
        )
        self.add_default_connections(self.get_default_connections_from_l1_generic_electrolyzer_controller())

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def get_default_connections_from_l1_generic_electrolyzer_controller(
        self,
    ) -> List[cp.ComponentConnection]:
        """Sets default connections of the controller in the Electroylzer."""

        connections: List[cp.ComponentConnection] = []
        controller_classname = controller_l1_electrolyzer.L1GenericElectrolyzerController.get_classname()
        connections.append(
            cp.ComponentConnection(
                GenericElectrolyzer.AvailableElectricity,
                controller_classname,
                controller_l1_electrolyzer.L1GenericElectrolyzerController.AvailableElectricity,
            )
        )
        return connections

    def i_save_state(self) -> None:
        """Save state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restore state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublecheck."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
        """Simulate."""
        if force_convergence:
            pass

        electricity_target = stsv.get_input_value(self.electricity_target_channel)
        if electricity_target < 0:
            raise ValueError("Target electricity needs to be positive in Electrolyzer")
        if 0 <= electricity_target < self.config.min_power:
            self.state.electricity = 0
            electricity_target = 0
        elif electricity_target > self.config.max_power:
            self.state.electricity = self.config.max_power
        else:
            self.state.electricity = electricity_target

        # interpolation between points
        if electricity_target == 0:
            hydrogen_output_in_kg_per_sec = 0.0
        else:
            hydrogen_output_in_kg_per_sec = self.config.min_hydrogen_production_rate + (
                (self.config.max_hydrogen_production_rate - self.config.min_hydrogen_production_rate)
                * (self.state.electricity - self.config.min_power)
                / (self.config.max_power - self.config.min_power)
            )
        self.state.hydrogen = hydrogen_output_in_kg_per_sec
        stsv.set_output_value(self.hydrogen_output_channel, self.state.hydrogen)
        stsv.set_output_value(self.electricity_output_channel, self.state.electricity)

    def write_to_report(self):
        """Writes the information of the current component to the report."""
        return self.config.get_string_dict()



""" Defines the simulation parameters class. This defines how the simulation will proceed. """
# clean
from __future__ import annotations
import os
import inspect
from typing import List, Optional
import enum

import datetime
from dataclasses import dataclass
from dataclass_wizard import JSONWizard

from hisim import log
from hisim.postprocessingoptions import PostProcessingOptions


@dataclass()
class SimulationParameters(JSONWizard):

    """Defines HOW the simulation is going to proceed: Time resolution, time span and all these things."""

    start_date: datetime.datetime
    end_date: datetime.datetime
    seconds_per_timestep: int
    post_processing_options: List[int]
    logging_level: int
    result_directory: str
    skip_finished_results: bool
    surplus_control: bool
    cache_dir_path: str
    multiple_buildings: bool

    def __init__(
        self,
        start_date: datetime.datetime,
        end_date: datetime.datetime,
        seconds_per_timestep: int,
        result_directory: str = "",
        post_processing_options: Optional[List[int]] = None,
        logging_level: int = log.LogPrio.INFORMATION,
        skip_finished_results: bool = False,
        surplus_control: bool = True,
        cache_dir_path: str = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), "inputs", "cache"),  # type: ignore
        multiple_buildings: bool = False,
    ):
        """Initializes the class."""
        self.start_date: datetime.datetime = start_date
        self.end_date: datetime.datetime = end_date
        self.seconds_per_timestep = seconds_per_timestep
        self.duration = end_date - start_date
        total_seconds = self.duration.total_seconds()
        self.timesteps: int = int(total_seconds / seconds_per_timestep)
        self.year: int = int(start_date.year)
        if post_processing_options is None:
            post_processing_options = []
        self.post_processing_options: List[int] = post_processing_options
        self.logging_level: int = logging_level  # Info # noqa
        self.result_directory: str = result_directory
        self.skip_finished_results: bool = skip_finished_results
        self.surplus_control = surplus_control
        self.cache_dir_path = cache_dir_path
        self.multiple_buildings = multiple_buildings
        self.figure_format = FigureFormat.PNG

    @classmethod
    def full_year(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a full year without any post processing, primarily for unit testing."""
        return cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year + 1, 1, 1),
            seconds_per_timestep,
            "",
        )

    def enable_all_options(self) -> None:
        """Enables all the post processing options ."""
        for option in PostProcessingOptions:
            self.post_processing_options.append(option)

    def enable_plots_only(self) -> None:
        """Enables line and carpet plots."""
        self.post_processing_options.append(PostProcessingOptions.PLOT_LINE)
        self.post_processing_options.append(PostProcessingOptions.PLOT_CARPET)
        # self.post_processing_options.append(PostProcessingOptions.PLOT_SANKEY)
        self.post_processing_options.append(PostProcessingOptions.PLOT_SINGLE_DAYS)
        self.post_processing_options.append(PostProcessingOptions.PLOT_MONTHLY_BAR_CHARTS)
        self.post_processing_options.append(PostProcessingOptions.OPEN_DIRECTORY_IN_EXPLORER)

    @classmethod
    def full_year_all_options(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a full year with all the post processing, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year + 1, 1, 1),
            seconds_per_timestep,
            "",
        )
        pars.enable_all_options()
        return pars

    @classmethod
    def full_year_with_only_plots(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a full year with all the post processing, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year + 1, 1, 1),
            seconds_per_timestep,
            "",
        )
        pars.enable_plots_only()
        return pars

    @classmethod
    def january_only_with_all_options(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a single january, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year, 1, 31),
            seconds_per_timestep,
            "",
        )
        pars.enable_all_options()
        return pars

    @classmethod
    def january_only_with_only_plots(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a single january, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year, 1, 31),
            seconds_per_timestep,
            "",
        )
        pars.enable_plots_only()
        return pars

    @classmethod
    def three_months_only(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a three months, primarily for unit testing."""
        return cls(
            datetime.datetime(year, 3, 1),
            datetime.datetime(year, 5, 31),
            seconds_per_timestep,
            "",
        )

    @classmethod
    def three_months_with_plots_only(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a three months, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 6, 1),
            datetime.datetime(year, 8, 31),
            seconds_per_timestep,
            "",
        )
        pars.enable_plots_only()
        return pars

    @classmethod
    def one_week_only(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a single week, primarily for unit testing."""
        return cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year, 1, 8),
            seconds_per_timestep,
            "",
        )

    @classmethod
    def one_week_with_only_plots(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a single week, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year, 1, 8),
            seconds_per_timestep,
            "",
        )

        pars.enable_plots_only()
        return pars

    @classmethod
    def one_day_only(cls, year: int, seconds_per_timestep: int = 60) -> SimulationParameters:
        """Generates a parameter set for a single day, primarily for unit testing."""
        return cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year, 1, 2),
            seconds_per_timestep,
            "",
        )

    @classmethod
    def one_day_only_with_all_options(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a single day, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year, 1, 2),
            seconds_per_timestep,
            "",
        )
        pars.enable_all_options()
        return pars

    @classmethod
    def one_day_only_with_only_plots(cls, year: int, seconds_per_timestep: int) -> SimulationParameters:
        """Generates a parameter set for a single day, primarily for unit testing."""
        pars = cls(
            datetime.datetime(year, 1, 1),
            datetime.datetime(year, 1, 2),
            seconds_per_timestep,
            "",
        )
        pars.enable_plots_only()
        return pars

    def get_unique_key(self) -> str:
        """Gets a unique key from a simulation parameter class."""
        return (
            str(self.start_date)
            + "###"
            + str(self.end_date)
            + "###"
            + str(self.seconds_per_timestep)
            + "###"
            + str(self.year)
            + "###"
            + str(self.timesteps)
        )

    def get_unique_key_as_list(self) -> List[str]:
        """Gets unique key from a simulation parameter class as list."""
        lines = []
        lines.append(f"Start date: {self.start_date}")
        lines.append(f"End date: {self.end_date}")
        lines.append(f"Simulation year: {self.year}")
        lines.append(f"Seconds per timestep: {self.seconds_per_timestep}")
        lines.append(f"Total number of timesteps: {self.timesteps}")
        return lines


class FigureFormat(str, enum.Enum):

    """Set Figure Formats."""

    PNG = ".png"
    JPG = ".jpg"



"""Generic battery."""

# clean

# Generic/Built-in
import copy
from typing import Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json

# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import utils
from hisim.components.generic_ev_charger import SimpleStorageState
from hisim.simulationparameters import SimulationParameters
from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum

__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


class GenericBatteryState:
    """Generic battery state class."""

    def __init__(
        self,
        init_stored_energy=0,
        max_stored_energy=None,
        min_stored_energy=None,
        max_var_stored_energy=None,
        min_var_stored_energy=None,
    ):
        """Initialize the class."""
        self.stored_energy = init_stored_energy
        self.max_stored_energy = max_stored_energy
        self.min_stored_energy = min_stored_energy
        self.max_var_stored_energy = max_var_stored_energy
        self.min_var_stored_energy = min_var_stored_energy
        self.charge_wh: float

    def charge(self, energy):
        """Charge."""
        energy = abs(energy)
        if self.stored_energy + energy < self.max_stored_energy:
            charge = energy
        else:
            charge = self.max_stored_energy - self.stored_energy

        if charge > self.max_var_stored_energy:
            charge = self.max_var_stored_energy

        self.stored_energy = charge + self.stored_energy
        self.charge_wh = charge

    def discharge(self, energy):
        """Discharge."""
        energy = -abs(energy)
        if self.stored_energy + energy > self.min_stored_energy:
            discharge = energy
        else:
            discharge = self.stored_energy - self.min_stored_energy

        if discharge < self.min_var_stored_energy:
            discharge = self.min_var_stored_energy

        self.stored_energy = discharge + self.stored_energy
        self.charge_wh = discharge


@dataclass_json
@dataclass
class GenericBatteryConfig(cp.ConfigBase):
    """Configuration of the Generic Battery."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return GenericBattery.get_full_classname()

    building_name: str
    name: str
    manufacturer: str
    model: str
    soc: float
    base: bool
    predictive: bool

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config."""
        return GenericBatteryConfig(
            building_name=building_name,
            name="Generic Battery",
            manufacturer="sonnen",
            model="sonnenBatterie 10 - 11,5 kWh",
            soc=10 / 15,
            base=False,
            predictive=False,
        )


@dataclass_json
@dataclass
class BatteryControllerConfig(cp.ConfigBase):
    """Configuration of the Generic Battery Controller."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return BatteryController.get_full_classname()

    building_name: str
    name: str

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default config."""
        return BatteryControllerConfig(
            building_name=building_name,
            name="Battery Controller",
        )


class GenericBattery(cp.Component):
    """Generic Battery class."""

    # Imports
    ElectricityInput = "ElectricityInput"
    State = "State"

    # Outputs
    StoredEnergy = "StoredEnergy"
    StateOfCharge = "StateOfCharge"
    ElectricityOutput = "ElectricityOutput"

    # simulation repository
    # MaximumBatteryCapacity = "MaximumBatteryCapacity"
    # MinimumBatteryCapacity = "MinimumBatteryCapacity"
    # MaximalChargingPower = "MaximalChargingPower"
    # MaximalDischargingPower = "MaximalDischargingPower"
    # BatteryEfficiency = "BatteryEfficiency"
    # InverterEfficiency = "InverterEfficiency"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: GenericBatteryConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.build(manufacturer=config.manufacturer, model=config.model, base=config.base)

        self.state = SimpleStorageState(
            max_var_val=self.max_var_stored_energy,
            min_var_val=self.min_var_stored_energy,
            stored_energy=self.max_stored_energy * config.soc,
        )
        self.previous_state = copy.deepcopy(self.state)

        self.input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityInput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )
        self.state_channel: cp.ComponentInput = self.add_input(
            self.component_name, self.State, lt.LoadTypes.ANY, lt.Units.ANY, True
        )

        self.state_of_charge_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.StateOfCharge,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description=f"here a description for {self.StateOfCharge} will follow.",
        )

        self.stored_energy_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.StoredEnergy,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT_HOUR,
            output_description=f"here a description for {self.StoredEnergy} will follow.",
        )

        self.electricity_output_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ElectricityOutput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            output_description=f"here a description for {self.ElectricityOutput} will follow.",
        )

    def build(self, manufacturer, model, base):
        """Build function."""
        self.base = base
        self.time_correction_factor = 1 / self.my_simulation_parameters.seconds_per_timestep
        self.seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep

        # Gets flexibilities, including heat pump
        battery_database = utils.load_smart_appliance("Battery")

        battery_found = False
        battery = None
        for battery in battery_database:
            if battery["Manufacturer"] == manufacturer and battery["Model"] == model:
                battery_found = True
                break

        if battery is None or not battery_found:
            raise Exception("Heat pump model not registered in the database")

        self.max_stored_energy = battery["Capacity"] * 1e3
        self.min_stored_energy = self.max_stored_energy * 0.0
        self.efficiency = battery["Efficiency"]
        self.efficiency_inverter = battery["Inverter Efficiency"]
        self.max_var_stored_energy = battery["Maximal Charging Power"] * 1e3 * self.time_correction_factor
        if "Maximal Discharging Power" in battery:
            self.min_var_stored_energy = -battery["Maximal Discharging Power"] * 1e3 * self.time_correction_factor
        else:
            self.min_var_stored_energy = -self.max_var_stored_energy

    def write_to_report(self):
        """Writes to report."""
        lines = []
        lines.append(f"MaxStoredEnergy: {self.max_stored_energy}")
        return lines

    # def i_save_state(self):
    #    self.previous_state = copy.copy(self.state)

    # def i_restore_state(self):
    #    self.state = copy.copy(self.previous_state)
    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = copy.deepcopy(self.state)

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        if self.config.predictive:
            # send battery specification to the mpc controller for planning the cost optimal operation
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.MAXIMUMBATTERYCAPACITY,
                entry=self.max_stored_energy,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.MINIMUMBATTERYCAPACITY,
                entry=self.min_stored_energy,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.MAXIMALCHARGINGPOWER,
                entry=self.max_var_stored_energy / self.time_correction_factor,
            )
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.MAXIMALDISCHARGINGPOWER,
                entry=-self.min_var_stored_energy / self.time_correction_factor,
            )
            SingletonSimRepository().set_entry(key=SingletonDictKeyEnum.BATTERYEFFICIENCY, entry=self.efficiency)
            SingletonSimRepository().set_entry(
                key=SingletonDictKeyEnum.INVERTEREFFICIENCY,
                entry=self.efficiency_inverter,
            )

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = copy.deepcopy(self.previous_state)

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        load = stsv.get_input_value(self.input_channel)
        state = stsv.get_input_value(self.state_channel)

        load = -load / self.seconds_per_timestep

        capacity = self.state.stored_energy
        max_capacity = self.max_stored_energy
        min_capacity = self.min_stored_energy

        if state == 1:
            charging_delta, after_capacity = self.state.store(
                max_capacity=max_capacity,
                current_capacity=capacity,
                val=load,
                efficiency=self.efficiency_inverter,
            )
        elif state == -1:
            charging_delta, after_capacity = self.state.withdraw(
                min_capacity=min_capacity,
                current_capacity=capacity,
                val=load,
                efficiency=self.efficiency_inverter * self.efficiency,
            )
        else:
            charging_delta = 0
            after_capacity = capacity

        stsv.set_output_value(self.state_of_charge_channel, after_capacity / max_capacity)
        stsv.set_output_value(self.stored_energy_channel, after_capacity)
        stsv.set_output_value(self.electricity_output_channel, charging_delta)


class BatteryController(cp.Component):
    """Battery Controller class."""

    ElectricityInput = "ElectricityInput"
    State = "State"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: BatteryControllerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ElectricityInput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            True,
        )
        self.state_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.State, lt.LoadTypes.ANY, lt.Units.ANY
        )

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        load = stsv.get_input_value(self.input_channel)
        state: float = 0
        if load < 0.0:
            state = 1
        elif load > 0.0:
            state = -1
        else:
            state = 0.0

        stsv.set_output_value(self.state_channel, state)
